# coding: utf-8
from flask import flash, render_template, url_for

from ...common import handle_metadata
from ..baseview import BaseView


metadata = dict(pageview=handle_metadata().get('pageview', 1))


class ServersView(BaseView):
    metadata = metadata

    def __init__(self):
        super(ServersView, self).__init__()

        self.opt = self.view_args['opt']
        self.project = self.view_args['project']
        self.version_job = self.view_args['version_job']
        self.spider = self.view_args['spider']

        self.url = 'http://%s/daemonstatus.json' % self.SCRAPYD_SERVER
        self.template = 'scrapydweb/servers.html'
        self.selected_nodes = []

    def dispatch_request(self, **kwargs):
        self.metadata['pageview'] += 1
        self.logger.debug('metadata: %s', self.metadata)

        if self.SCRAPYD_SERVERS_AMOUNT > 1 and not (self.metadata['pageview'] > 2 and self.metadata['pageview'] % 100):
            if not self.ENABLE_AUTH:
                flash("Set 'ENABLE_AUTH = True' to enable basic auth for web UI", self.INFO)
            if self.IS_LOCAL_SCRAPYD_SERVER and not self.ENABLE_LOGPARSER:
                flash("Set 'ENABLE_LOGPARSER = True' to run LogParser as a subprocess at startup", self.WARN)
            if not self.ENABLE_MONITOR:
                flash("Set 'ENABLE_MONITOR = True' to enable the monitor feature", self.INFO)

        if self.POST:
            self.selected_nodes = self.get_selected_nodes()
        else:
            if self.SCRAPYD_SERVERS_AMOUNT == 1:
                self.selected_nodes = [1]
            else:
                self.selected_nodes = []

        kwargs = dict(
            node=self.node,
            opt=self.opt,
            project=self.project,
            version_job=self.version_job,
            spider=self.spider,
            url=self.url,
            selected_nodes=self.selected_nodes,
            IS_IE_EDGE=self.IS_IE_EDGE,
            pageview=self.metadata['pageview'],
            FEATURES=self.FEATURES,
            DEFAULT_LATEST_VERSION=self.DEFAULT_LATEST_VERSION,
            url_daemonstatus=url_for('api', node=self.node, opt='daemonstatus'),
            url_getreports=url_for('clusterreports', node=self.node, project='PROJECT_PLACEHOLDER',
                                   spider='SPIDER_PLACEHOLDER', job='JOB_PLACEHOLDER'),
            url_liststats=url_for('api', node=self.node, opt='liststats', project='PROJECT_PLACEHOLDER',
                                  version_spider_job='JOB_PLACEHOLDER'),
            url_listprojects=url_for('api', node=self.node, opt='listprojects'),
            url_listversions=url_for('api', node=self.node, opt='listversions', project='PROJECT_PLACEHOLDER'),
            url_listspiders=url_for('api', node=self.node, opt='listspiders', project='PROJECT_PLACEHOLDER',
                                    version_spider_job='VERSION_PLACEHOLDER'),
            url_listjobs=url_for('api', node=self.node, opt='listjobs', project='PROJECT_PLACEHOLDER'),
            url_deploy=url_for('deploy', node=self.node),
            url_schedule=url_for('schedule', node=self.node, project='PROJECT_PLACEHOLDER',
                                 version='VERSION_PLACEHOLDER', spider='SPIDER_PLACEHOLDER'),
            url_stop=url_for('multinode', node=self.node, opt='stop', project='PROJECT_PLACEHOLDER',
                             version_job='JOB_PLACEHOLDER'),
            url_delversion=url_for('multinode', node=self.node, opt='delversion', project='PROJECT_PLACEHOLDER',
                                   version_job='VERSION_PLACEHOLDER'),
            url_delproject=url_for('multinode', node=self.node, opt='delproject', project='PROJECT_PLACEHOLDER')
        )
        return render_template(self.template, **kwargs)



# coding: utf-8

__title__ = 'scrapydwebx'
__version__ = '1.5.1'
__author__ = 'payala'
__author_email__ = 'ppayala@gmail.com'
__url__ = 'https://github.com/payala/scrapydwebx'
__license__ = 'GNU General Public License v3.0'
__description__ = ("Web app for Scrapyd cluster management, "
                   "with support for Scrapy log analysis & visualization.")



# coding: utf-8
import logging
from multiprocessing.dummy import Pool as ThreadPool
import os
import re

from ..common import handle_metadata, handle_slash, json_dumps, session
from ..models import create_jobs_table, db
from ..utils.scheduler import scheduler
from ..utils.setup_database import test_database_url_pattern
from ..vars import (ALLOWED_SCRAPYD_LOG_EXTENSIONS, ALERT_TRIGGER_KEYS,
                    SCHEDULER_STATE_DICT, STATE_PAUSED, STATE_RUNNING,
                    SCHEDULE_ADDITIONAL, STRICT_NAME_PATTERN, UA_DICT,
                    jobs_table_map)
from .send_email import send_email
from .sub_process import init_logparser, init_poll


logger = logging.getLogger(__name__)

REPLACE_URL_NODE_PATTERN = re.compile(r'(:\d+/)\d+/')
EMAIL_PATTERN = re.compile(r'^[^@]+@[^@]+\.[^@]+$')
HASH = '#' * 100
# r'^(?:(?:(.*?)\:)(?:(.*?)@))?(.*?)(?:\:(.*?))?(?:#(.*?))?$'
SCRAPYD_SERVER_PATTERN = re.compile(r"""
                                        ^
                                        (?:
                                            (?:(.*?):)      # username:
                                            (?:(.*?)@)      # password@
                                        )?
                                        (.*?)               # ip
                                        (?::(.*?))?         # :port
                                        (?:\#(.*?))?        # #group
                                        $
                                    """, re.X)


def check_app_config(config):
    def check_assert(key, default, is_instance, allow_zero=True, non_empty=False, containing_type=None):
        if is_instance is int:
            if allow_zero:
                should_be = "a non-negative integer"
            else:
                should_be = "a positive integer"
        else:
            should_be = "an instance of %s%s" % (is_instance, ' and not empty' if non_empty else '')

        value = config.setdefault(key, default)
        kws = dict(
            key=key,
            should_be=should_be,
            containing_type=', containing elements of type %s' % containing_type if containing_type else '',
            value="'%s'" % value if isinstance(value, str) else value
        )
        to_assert = u"{key} should be {should_be}{containing_type}. Current value: {value}".format(**kws)

        assert (isinstance(value, is_instance)
                and (not isinstance(value, bool) if is_instance is int else True)  # isinstance(True, int) => True
                and (value > (-1 if allow_zero else 0) if is_instance is int else True)
                and (value if non_empty else True)
                and (all([isinstance(i, containing_type) for i in value]) if containing_type else True)), to_assert

    logger.debug("Checking app config")

    # ScrapydWeb
    check_assert('SCRAPYDWEB_BIND', '0.0.0.0', str, non_empty=True)
    SCRAPYDWEB_PORT = config.setdefault('SCRAPYDWEB_PORT', 5000)
    try:
        assert not isinstance(SCRAPYDWEB_PORT, bool)
        SCRAPYDWEB_PORT = int(SCRAPYDWEB_PORT)
        assert SCRAPYDWEB_PORT > 0
    except (TypeError, ValueError, AssertionError):
        assert False, "SCRAPYDWEB_PORT should be a positive integer. Current value: %s" % SCRAPYDWEB_PORT

    check_assert('ENABLE_AUTH', False, bool)
    if config.get('ENABLE_AUTH', False):
        # May be int from config file
        check_assert('USERNAME', '', str, non_empty=True)
        check_assert('PASSWORD', '', str, non_empty=True)
        handle_metadata('username', config['USERNAME'])
        handle_metadata('password', config['PASSWORD'])
        logger.info("Basic auth enabled with USERNAME/PASSWORD: '%s'/'%s'", config['USERNAME'], config['PASSWORD'])

    check_assert('ENABLE_HTTPS', False, bool)
    if config.get('ENABLE_HTTPS', False):
        logger.info("HTTPS mode enabled: ENABLE_HTTPS = %s", config['ENABLE_HTTPS'])
        for k in ['CERTIFICATE_FILEPATH', 'PRIVATEKEY_FILEPATH']:
            check_assert(k, '', str, non_empty=True)
            assert os.path.isfile(config[k]), "%s not found: %s" % (k, config[k])
        logger.info("Running in HTTPS mode: %s, %s", config['CERTIFICATE_FILEPATH'], config['PRIVATEKEY_FILEPATH'])

    _protocol = 'https' if config.get('ENABLE_HTTPS', False) else 'http'
    _bind = config.get('SCRAPYDWEB_BIND', '0.0.0.0')
    _bind = '127.0.0.1' if _bind == '0.0.0.0' else _bind
    config['URL_SCRAPYDWEB'] = '%s://%s:%s' % (_protocol, _bind, config.get('SCRAPYDWEB_PORT', 5000))
    handle_metadata('url_scrapydweb', config['URL_SCRAPYDWEB'])
    logger.info("Setting up URL_SCRAPYDWEB: %s", config['URL_SCRAPYDWEB'])

    # Scrapy
    check_assert('SCRAPY_PROJECTS_DIR', '', str)
    SCRAPY_PROJECTS_DIR = config.get('SCRAPY_PROJECTS_DIR', '')
    if SCRAPY_PROJECTS_DIR:
        assert os.path.isdir(SCRAPY_PROJECTS_DIR), "SCRAPY_PROJECTS_DIR not found: %s" % SCRAPY_PROJECTS_DIR
        logger.info("Setting up SCRAPY_PROJECTS_DIR: %s", handle_slash(SCRAPY_PROJECTS_DIR))

    # Scrapyd
    check_scrapyd_servers(config)
    # For JobsView
    for node, scrapyd_server in enumerate(config['SCRAPYD_SERVERS'], 1):
        # Note that check_app_config() is executed multiple times in test
        if node not in jobs_table_map:
            jobs_table_map[node] = create_jobs_table(re.sub(STRICT_NAME_PATTERN, '_', scrapyd_server))
    db.create_all(bind='jobs')
    logger.debug("Created %s tables for JobsView", len(jobs_table_map))

    check_assert('LOCAL_SCRAPYD_LOGS_DIR', '', str)
    check_assert('LOCAL_SCRAPYD_SERVER', '', str)
    LOCAL_SCRAPYD_LOGS_DIR = config.get('LOCAL_SCRAPYD_LOGS_DIR', '')
    if LOCAL_SCRAPYD_LOGS_DIR:
        assert os.path.isdir(LOCAL_SCRAPYD_LOGS_DIR), "LOCAL_SCRAPYD_LOGS_DIR not found: %s" % LOCAL_SCRAPYD_LOGS_DIR
        logger.info("Setting up LOCAL_SCRAPYD_LOGS_DIR: %s", handle_slash(LOCAL_SCRAPYD_LOGS_DIR))
        LOCAL_SCRAPYD_SERVER = config.get('LOCAL_SCRAPYD_SERVER', '')
        if LOCAL_SCRAPYD_SERVER and not re.search(r':\d+$', LOCAL_SCRAPYD_SERVER):
            LOCAL_SCRAPYD_SERVER += ':6800'
            config['LOCAL_SCRAPYD_SERVER'] = LOCAL_SCRAPYD_SERVER
        if len(config['SCRAPYD_SERVERS']) > 1:
            assert LOCAL_SCRAPYD_SERVER, \
                ("The LOCAL_SCRAPYD_SERVER option must be set up since you have added multiple Scrapyd servers "
                 "and set up the LOCAL_SCRAPYD_LOGS_DIR option.\nOtherwise, just set LOCAL_SCRAPYD_LOGS_DIR to ''")
        else:
            if not LOCAL_SCRAPYD_SERVER:
                config['LOCAL_SCRAPYD_SERVER'] = config['SCRAPYD_SERVERS'][0]
                LOCAL_SCRAPYD_SERVER = config['LOCAL_SCRAPYD_SERVER']
        logger.info("Setting up LOCAL_SCRAPYD_SERVER: %s", LOCAL_SCRAPYD_SERVER)
        assert LOCAL_SCRAPYD_SERVER in config['SCRAPYD_SERVERS'], \
            "LOCAL_SCRAPYD_SERVER '%s' is not in the Scrapyd servers you have added:\n%s" % (
                LOCAL_SCRAPYD_SERVER, config['SCRAPYD_SERVERS'])
    # else:
    #     _path = os.path.join(os.path.expanduser('~'), 'logs')
    #     if os.path.isdir(_path):
    #         config['LOCAL_SCRAPYD_LOGS_DIR'] = _path
    #         logger.info("Found LOCAL_SCRAPYD_LOGS_DIR: %s", config['LOCAL_SCRAPYD_LOGS_DIR'])

    check_assert('SCRAPYD_LOG_EXTENSIONS', ALLOWED_SCRAPYD_LOG_EXTENSIONS, list, non_empty=True, containing_type=str)
    SCRAPYD_LOG_EXTENSIONS = config.get('SCRAPYD_LOG_EXTENSIONS', ALLOWED_SCRAPYD_LOG_EXTENSIONS)
    assert all([not i or i.startswith('.') for i in SCRAPYD_LOG_EXTENSIONS]), \
        ("SCRAPYD_LOG_EXTENSIONS should be a list like %s. "
         "Current value: %s" % (ALLOWED_SCRAPYD_LOG_EXTENSIONS, SCRAPYD_LOG_EXTENSIONS))
    logger.info("Locating scrapy logfiles with SCRAPYD_LOG_EXTENSIONS: %s", SCRAPYD_LOG_EXTENSIONS)

    # LogParser
    check_assert('ENABLE_LOGPARSER', False, bool)
    if config.get('ENABLE_LOGPARSER', False):
        assert config.get('LOCAL_SCRAPYD_LOGS_DIR', ''), \
            ("In order to automatically run LogParser at startup, you have to set up the LOCAL_SCRAPYD_LOGS_DIR option "
             "first.\nOtherwise, set 'ENABLE_LOGPARSER = False' if you are not running any Scrapyd service "
             "on the current ScrapydWeb host.\nNote that you can run the LogParser service separately "
             "via command 'logparser' as you like. ")
    check_assert('BACKUP_STATS_JSON_FILE', True, bool)

    # Run Spider
    check_assert('SCHEDULE_EXPAND_SETTINGS_ARGUMENTS', False, bool)
    check_assert('SCHEDULE_CUSTOM_USER_AGENT', '', str)
    config['SCHEDULE_CUSTOM_USER_AGENT'] = config['SCHEDULE_CUSTOM_USER_AGENT'] or 'Mozilla/5.0'
    UA_DICT.update(custom=config['SCHEDULE_CUSTOM_USER_AGENT'])
    if config.get('SCHEDULE_USER_AGENT', None) is not None:
        check_assert('SCHEDULE_USER_AGENT', '', str)
        user_agent = config['SCHEDULE_USER_AGENT']
        assert user_agent in UA_DICT.keys(), \
            "SCHEDULE_USER_AGENT should be any value of %s. Current value: %s" % (UA_DICT.keys(), user_agent)
    if config.get('SCHEDULE_ROBOTSTXT_OBEY', None) is not None:
        check_assert('SCHEDULE_ROBOTSTXT_OBEY', False, bool)
    if config.get('SCHEDULE_COOKIES_ENABLED', None) is not None:
        check_assert('SCHEDULE_COOKIES_ENABLED', False, bool)
    if config.get('SCHEDULE_CONCURRENT_REQUESTS', None) is not None:
        check_assert('SCHEDULE_CONCURRENT_REQUESTS', 16, int, allow_zero=False)
    if config.get('SCHEDULE_DOWNLOAD_DELAY', None) is not None:
        download_delay = config['SCHEDULE_DOWNLOAD_DELAY']
        if isinstance(download_delay, float):
            assert download_delay >= 0.0, \
                "SCHEDULE_DOWNLOAD_DELAY should a non-negative number. Current value: %s" % download_delay
        else:
            check_assert('SCHEDULE_DOWNLOAD_DELAY', 0, int)
    check_assert('SCHEDULE_ADDITIONAL', SCHEDULE_ADDITIONAL, str)

    # Page Display
    check_assert('SHOW_SCRAPYD_ITEMS', True, bool)
    check_assert('SHOW_JOBS_JOB_COLUMN', False, bool)
    check_assert('JOBS_FINISHED_JOBS_LIMIT', 0, int)
    check_assert('JOBS_RELOAD_INTERVAL', 300, int)
    check_assert('DAEMONSTATUS_REFRESH_INTERVAL', 10, int)

    # Send text
    check_assert('SLACK_TOKEN', '', str)
    check_assert('SLACK_CHANNEL', '', str)
    config['SLACK_CHANNEL'] = config['SLACK_CHANNEL'] or 'general'

    check_assert('TELEGRAM_TOKEN', '', str)
    check_assert('TELEGRAM_CHAT_ID', 0, int)

    check_assert('EMAIL_PASSWORD', '', str)
    if config.get('EMAIL_PASSWORD', ''):
        logger.debug("Found EMAIL_PASSWORD, checking email settings")
        check_assert('EMAIL_SUBJECT', '', str)
        check_assert('EMAIL_USERNAME', '', str)  # '' would default to config['EMAIL_SENDER']
        # check_assert('EMAIL_PASSWORD', '', str, non_empty=True)
        check_assert('EMAIL_SENDER', '', str, non_empty=True)
        EMAIL_SENDER = config['EMAIL_SENDER']
        assert re.search(EMAIL_PATTERN, EMAIL_SENDER), \
            "EMAIL_SENDER should contain '@', like 'username@gmail.com'. Current value: %s" % EMAIL_SENDER
        check_assert('EMAIL_RECIPIENTS', [], list, non_empty=True, containing_type=str)
        EMAIL_RECIPIENTS = config['EMAIL_RECIPIENTS']
        assert all([re.search(EMAIL_PATTERN, i) for i in EMAIL_RECIPIENTS]), \
            ("All elements in EMAIL_RECIPIENTS should contain '@', like 'username@gmail.com'. "
             "Current value: %s") % EMAIL_RECIPIENTS
        if not config.get('EMAIL_USERNAME', ''):
            config['EMAIL_USERNAME'] = config['EMAIL_SENDER']

        check_assert('SMTP_SERVER', '', str, non_empty=True)
        check_assert('SMTP_PORT', 0, int, allow_zero=False)
        check_assert('SMTP_OVER_SSL', False, bool)
        check_assert('SMTP_CONNECTION_TIMEOUT', 30, int, allow_zero=False)

    # Monitor & Alert
    check_assert('ENABLE_MONITOR', False, bool)
    if config.get('ENABLE_MONITOR', False):
        check_assert('POLL_ROUND_INTERVAL', 300, int, allow_zero=False)
        check_assert('POLL_REQUEST_INTERVAL', 10, int, allow_zero=False)

        check_assert('ENABLE_SLACK_ALERT', False, bool)
        check_assert('ENABLE_TELEGRAM_ALERT', False, bool)
        check_assert('ENABLE_EMAIL_ALERT', False, bool)

        # For compatibility with Python 3 using range()
        try:
            config['ALERT_WORKING_DAYS'] = list(config.get('ALERT_WORKING_DAYS', []))
        except TypeError:
            pass
        check_assert('ALERT_WORKING_DAYS', [], list, non_empty=True, containing_type=int)
        ALERT_WORKING_DAYS = config['ALERT_WORKING_DAYS']
        assert all([not isinstance(i, bool) and i in range(1, 8) for i in ALERT_WORKING_DAYS]), \
            "Element in ALERT_WORKING_DAYS should be between 1 and 7. Current value: %s" % ALERT_WORKING_DAYS

        try:
            config['ALERT_WORKING_HOURS'] = list(config.get('ALERT_WORKING_HOURS', []))
        except TypeError:
            pass
        check_assert('ALERT_WORKING_HOURS', [], list, non_empty=True, containing_type=int)
        ALERT_WORKING_HOURS = config['ALERT_WORKING_HOURS']
        assert all([not isinstance(i, bool) and i in range(24) for i in ALERT_WORKING_HOURS]), \
            "Element in ALERT_WORKING_HOURS should be between 0 and 23. Current value: %s" % ALERT_WORKING_HOURS

        check_assert('ON_JOB_RUNNING_INTERVAL', 0, int)
        check_assert('ON_JOB_FINISHED', False, bool)

        for k in ALERT_TRIGGER_KEYS:
            check_assert('LOG_%s_THRESHOLD' % k, 0, int)
            check_assert('LOG_%s_TRIGGER_STOP' % k, False, bool)
            check_assert('LOG_%s_TRIGGER_FORCESTOP' % k, False, bool)

        if config.get('ENABLE_SLACK_ALERT', False):
            check_assert('SLACK_TOKEN', '', str, non_empty=True)
            check_slack_telegram(config, service='slack')
        if config.get('ENABLE_TELEGRAM_ALERT', False):
            check_assert('TELEGRAM_TOKEN', '', str, non_empty=True)
            check_assert('TELEGRAM_CHAT_ID', 0, int, allow_zero=False)
            check_slack_telegram(config, service='telegram')
        if config.get('ENABLE_EMAIL_ALERT', False):
            check_assert('EMAIL_PASSWORD', '', str, non_empty=True)
            check_email(config)

    # System
    check_assert('DEBUG', False, bool)
    check_assert('VERBOSE', False, bool)
    # if config.get('VERBOSE', False):
        # logging.getLogger('apscheduler').setLevel(logging.DEBUG)
    # else:
        # logging.getLogger('apscheduler').setLevel(logging.WARNING)
    check_assert('DATA_PATH', '', str)
    check_assert('DATABASE_URL', '', str)
    database_url = config.get('DATABASE_URL', '')
    if database_url:
        assert any(test_database_url_pattern(database_url)), "Invalid format of DATABASE_URL: %s" % database_url

    # Apscheduler
    # In __init__.py create_app(): scheduler.start(paused=True)
    if handle_metadata().get('scheduler_state', STATE_RUNNING) != STATE_PAUSED:
        scheduler.resume()
    logger.info("Scheduler for timer tasks: %s", SCHEDULER_STATE_DICT[scheduler.state])

    check_assert('JOBS_SNAPSHOT_INTERVAL', 300, int)
    JOBS_SNAPSHOT_INTERVAL = config.get('JOBS_SNAPSHOT_INTERVAL', 300)
    if JOBS_SNAPSHOT_INTERVAL:
        # TODO: with app.app_context(): url = url_for('jobs', node=1)
        # Working outside of application context.
        # only because before app.run?!
        username = config.get('USERNAME', '')
        password = config.get('PASSWORD', '')
        kwargs = dict(
            # 'http(s)://127.0.0.1:5000' + '/1/jobs/'
            url_jobs=config['URL_SCRAPYDWEB'] + handle_metadata().get('url_jobs', '/1/jobs/'),
            auth=(username, password) if username and password else None,
            nodes=list(range(1, len(config['SCRAPYD_SERVERS']) + 1))
        )
        logger.info(scheduler.add_job(id='jobs_snapshot', replace_existing=True,
                                      func=create_jobs_snapshot, args=None, kwargs=kwargs,
                                      trigger='interval', seconds=JOBS_SNAPSHOT_INTERVAL,
                                      misfire_grace_time=60, coalesce=True, max_instances=1, jobstore='memory'))

    # Subprocess
    init_subprocess(config)


def create_jobs_snapshot(url_jobs, auth, nodes):
    for node in nodes:
        url_jobs = re.sub(REPLACE_URL_NODE_PATTERN, r'\g<1>%s/' % node, url_jobs, count=1)
        try:
            r = session.post(url_jobs, auth=auth, timeout=60)
            assert r.status_code == 200, "Request got status_code: %s" % r.status_code
        except Exception as err:
            print("Fail to create jobs snapshot: %s\n%s" % (url_jobs, err))
        # else:
        #     print(url_jobs, r.status_code)


def check_scrapyd_servers(config):
    SCRAPYD_SERVERS = config.get('SCRAPYD_SERVERS', []) or ['127.0.0.1:6800']
    SCRAPYD_SERVERS_PUBLIC_URLS = config.get('SCRAPYD_SERVERS_PUBLIC_URLS', None) or [''] * len(SCRAPYD_SERVERS)
    assert len(SCRAPYD_SERVERS_PUBLIC_URLS) == len(SCRAPYD_SERVERS), \
        "SCRAPYD_SERVERS_PUBLIC_URLS should have same length with SCRAPYD_SERVERS:\n%s\nvs.\n%s" % (
        SCRAPYD_SERVERS_PUBLIC_URLS, SCRAPYD_SERVERS)

    servers = []
    for idx, (server, public_url) in enumerate(zip(SCRAPYD_SERVERS, SCRAPYD_SERVERS_PUBLIC_URLS)):
        if isinstance(server, tuple):
            assert len(server) == 5, ("Scrapyd server should be a tuple of 5 elements, "
                                      "current value: %s" % str(server))
            usr, psw, ip, port, group = server
        else:
            usr, psw, ip, port, group = re.search(SCRAPYD_SERVER_PATTERN, server.strip()).groups()
        ip = ip.strip() if ip and ip.strip() else '127.0.0.1'
        port = port.strip() if port and port.strip() else '6800'
        group = group.strip() if group and group.strip() else ''
        auth = (usr, psw) if usr and psw else None
        public_url = public_url.strip(' /')
        servers.append((group, ip, port, auth, public_url))

    def key_func(arg):
        _group, _ip, _port, _auth, _public_url = arg
        parts = _ip.split('.')
        parts = [('0' * (3 - len(part)) + part) for part in parts]
        return [_group, '.'.join(parts), int(_port)]

    servers = sorted(set(servers), key=key_func)
    check_scrapyd_connectivity(servers)

    config['SCRAPYD_SERVERS'] = ['%s:%s' % (ip, port) for (group, ip, port, auth, public_url) in servers]
    config['SCRAPYD_SERVERS_GROUPS'] = [group for (group, ip, port, auth, public_url) in servers]
    config['SCRAPYD_SERVERS_AUTHS'] = [auth for (group, ip, port, auth, public_url) in servers]
    config['SCRAPYD_SERVERS_PUBLIC_URLS'] = [public_url for (group, ip, port, auth, public_url) in servers]


def check_scrapyd_connectivity(servers):
    logger.debug("Checking connectivity of SCRAPYD_SERVERS...")

    def check_connectivity(server):
        (_group, _ip, _port, _auth, _public_url) = server
        try:
            url = 'http://%s:%s' % (_ip, _port)
            r = session.get(url, auth=_auth, timeout=10)
            assert r.status_code == 200, "%s got status_code %s" % (url, r.status_code)
        except Exception as err:
            logger.error(err)
            return False
        else:
            return True

    # with ThreadPool(min(len(servers), 100)) as pool:  # Works in python 3.3 and up
        # results = pool.map(check_connectivity, servers)
    pool = ThreadPool(min(len(servers), 100))
    results = pool.map(check_connectivity, servers)
    pool.close()
    pool.join()

    print("\nIndex {group:<20} {server:<21} Connectivity Auth".format(
          group='Group', server='Scrapyd IP:Port'))
    print(HASH)
    for idx, ((group, ip, port, auth, public_url), result) in enumerate(zip(servers, results), 1):
        print("{idx:_<5} {group:_<20} {server:_<22} {result:_<11} {auth}".format(
              idx=idx, group=group or 'None', server='%s:%s' % (ip, port), auth=auth, result=str(result)))
    print(HASH + '\n')

    assert any(results), "None of your SCRAPYD_SERVERS could be connected. "


def check_slack_telegram(config, service):
    logger.debug("Trying to send %s..." % service)
    text = '%s alert enabled #scrapydweb' % service.capitalize()
    alert = "Fail to send text via %s, you may need to set 'ENABLE_%s_ALERT = False'" % (
        service.capitalize(), service.upper())
    if service == 'slack':
        url = 'https://slack.com/api/chat.postMessage'
        data = dict(token=config['SLACK_TOKEN'], channel=config['SLACK_CHANNEL'], text=text)
    else:
        url = 'https://api.telegram.org/bot%s/sendMessage' % config['TELEGRAM_TOKEN']
        data = dict(chat_id=config['TELEGRAM_CHAT_ID'], text=text)
    r = None
    try:
        r = session.post(url, data=data, timeout=30)
        js = r.json()
        assert r.status_code == 200 and js['ok'] is True
    except Exception as err:
        logger.error(err)
        logger.error("url: %s", url)
        logger.error("data: %s", data)
        if r is not None:
            logger.error("status_code: %s", r.status_code)
            logger.error("response: %s", r.text)
        assert False, alert
    else:
        logger.info(text)


def check_email(config):
    kwargs = dict(
        email_username=config['EMAIL_USERNAME'],
        email_password=config['EMAIL_PASSWORD'],
        email_sender=config['EMAIL_SENDER'],
        email_recipients=config['EMAIL_RECIPIENTS'],
        smtp_server=config['SMTP_SERVER'],
        smtp_port=config['SMTP_PORT'],
        smtp_over_ssl=config.get('SMTP_OVER_SSL', False),
        smtp_connection_timeout=config.get('SMTP_CONNECTION_TIMEOUT', 30),
    )
    kwargs['to_retry'] = True
    kwargs['subject'] = 'Email alert enabled #scrapydweb'
    kwargs['content'] = json_dumps(dict(EMAIL_SENDER=config['EMAIL_SENDER'],
                                        EMAIL_RECIPIENTS=config['EMAIL_RECIPIENTS']))

    logger.debug("Trying to send email (smtp_connection_timeout=%s)...", config.get('SMTP_CONNECTION_TIMEOUT', 30))
    result, reason = send_email(**kwargs)
    if not result and os.environ.get('TEST_ON_CIRCLECI', 'False') == 'False':
        logger.debug("kwargs for send_email():\n%s", json_dumps(kwargs, sort_keys=False))
    assert result, "Fail to send email. Modify the email settings above or set 'ENABLE_EMAIL_ALERT = False'"
    logger.info(kwargs['subject'])


def init_subprocess(config):
    if config.get('ENABLE_LOGPARSER', False):
        config['LOGPARSER_PID'] = init_logparser(config)
    else:
        config['LOGPARSER_PID'] = None
    handle_metadata('logparser_pid', config['LOGPARSER_PID'])

    if config.get('ENABLE_MONITOR', False):
        config['POLL_PID'] = init_poll(config)
    else:
        config['POLL_PID'] = None
    handle_metadata('poll_pid', config['POLL_PID'])



# coding: utf-8
from collections import OrderedDict

from flask import url_for

from tests.utils import cst, get_text, req, upload_file_deploy


def test_listprojects(app, client):
    # upload_file_deploy(app, client, filename='demo.egg', project=cst.PROJECT, redirect_project=cst.PROJECT)
    upload_file_deploy(app, client, filename='demo.zip', project=cst.PROJECT, redirect_project=cst.PROJECT)

    req(app, client, view='projects', kws=dict(node=1), ins='Get the list of projects uploaded')


def test_listversions(app, client):
    with app.test_request_context():
        url_delproject = url_for('projects', node=1, opt='delproject', project=cst.PROJECT)
        req(app, client, view='projects', kws=dict(node=1, opt='listversions', project=cst.PROJECT),
            ins=['Delete Project', 'Delete Version', url_delproject])

        # {"status": "ok", "versions": []}
        url_delproject = url_for('projects', node=1, opt='delproject', project=cst.FAKE_PROJECT)
        req(app, client, view='projects', kws=dict(node=1, opt='listversions', project=cst.FAKE_PROJECT),
            ins=['Delete Project', url_delproject], nos='Delete the version')


# test_listspiders
# test_delversion
# test_delproject
def test_listspiders_del(app, client):
    with app.test_request_context():
        d = OrderedDict()  # For python 2 compatibility

        d['listspiders'] = dict(
            url=url_for('projects', node=1, opt='listspiders', project=cst.PROJECT, version_spider_job=cst.VERSION),
            checks=['Run Spider (test)']
        )
        d['listspiders_fail'] = dict(
            url=url_for('projects', node=1, opt='listspiders',
                        project=cst.FAKE_PROJECT, version_spider_job=cst.FAKE_VERSION),
            checks=['listspiders.json', 'No such file or directory']
        )

        d['delversion'] = dict(
            url=url_for('projects', node=1, opt='delversion', project=cst.PROJECT, version_spider_job=cst.VERSION),
            checks=['version deleted']
        )
        d['delversion_fail'] = dict(
            url=url_for('projects', node=1, opt='delversion',
                        project=cst.FAKE_PROJECT, version_spider_job=cst.FAKE_VERSION),
            checks=['delversion.json', 'See details below']
        )

        d['delproject'] = dict(
            url=url_for('projects', node=1, opt='delproject', project=cst.PROJECT),
            checks=['project deleted']
        )
        d['delproject_fail'] = dict(
            url=url_for('projects', node=1, opt='delproject', project=cst.FAKE_PROJECT),
            checks=['delproject.json', 'See details below']
        )

        for k, v in d.items():
            if k == 'delproject':  # Should use OrderedDict For python 2 compatibility
                upload_file_deploy(app, client, filename='demo.zip', project=cst.PROJECT, redirect_project=cst.PROJECT)
            response = client.get(v['url'])
            text = get_text(response)
            for c in v['checks']:
                assert c in text



# coding: utf-8
import datetime
import json

from flask import render_template, request, url_for

from ..baseview import BaseView


class ProjectsView(BaseView):

    def __init__(self):
        super(ProjectsView, self).__init__()

        self.opt = self.view_args['opt']
        self.project = self.view_args['project']
        self.version_spider_job = self.view_args['version_spider_job']

        self.text = ''
        self.js = {}

    def dispatch_request(self, **kwargs):
        # self.text = api(self.node, self.opt, self.project, self.version_spider_job)
        _url = url_for('api', node=self.node, opt=self.opt, project=self.project,
                       version_spider_job=self.version_spider_job)  # '/1/api/listprojects/'
        self.text = self.get_response_from_view(_url, as_json=False)
        # _bind = '127.0.0.1' if self.SCRAPYDWEB_BIND == '0.0.0.0' else self.SCRAPYDWEB_BIND
        # _url = 'http://{}:{}{}'.format(_bind, self.SCRAPYDWEB_PORT, _url)
        # _auth = (self.USERNAME, self.PASSWORD) if self.ENABLE_AUTH else None
        # status_code, self.text = self.make_request(_url, auth=_auth, as_json=False)
        self.js = json.loads(self.text)

        if self.js['status'] == self.OK:
            return getattr(self, self.opt)()
        else:
            return self.handle_status_error()

    @staticmethod
    def delproject():
        return '<em class="pass">project deleted</em>'

    @staticmethod
    def delversion():
        return '<em class="pass">version deleted</em>'

    def handle_status_error(self):
        if self.opt == 'listversions':
            kwargs = dict(
                url=self.js['url'],
                status=self.js['status'],
                url_deploy=url_for('deploy', node=self.node),
                url_delproject=url_for('projects', node=self.node, opt='delproject', project=self.project),
                project=self.project,
                text=self.text,
                tip=self.js.get('tip', '')
            )
            return render_template('scrapydweb/listversions_error.html', **kwargs)
        else:
            if self.POST:
                # Pass request.url instead of js['url'], for GET method
                return ('<a class="request" target="_blank" href="%s">REQUEST</a>'
                        '<em class="fail"> got status: %s</em>') % (request.url, self.js['status'])
            else:
                alert = 'REQUEST got status: %s' % self.js['status']
                message = self.js.get('message', '')
                if message:
                    self.js['message'] = 'See details below'
                return render_template(self.template_fail, node=self.node,
                                       alert=alert, text=self.json_dumps(self.js), message=message)

    def listprojects(self):
        results = []
        for project in self.js['projects']:
            url_listversions = url_for('projects', node=self.node, opt='listversions', project=project)
            results.append((project, url_listversions))

        kwargs = dict(
            node=self.node,
            url=self.js['url'],
            node_name=self.js['node_name'],
            results=results,
            url_deploy=url_for('deploy', node=self.node)
        )
        return render_template('scrapydweb/projects.html', **kwargs)

    def listspiders(self):
        spiders = self.js['spiders']
        results = []
        for spider in spiders:
            url_schedule = url_for('schedule', node=self.node,
                                   project=self.project, version=self.version_spider_job, spider=spider)
            url_multinode_schedule = url_for('servers', node=self.node, opt='schedule',
                                             project=self.project, version_job=self.version_spider_job, spider=spider)
            results.append((spider, url_schedule, url_multinode_schedule))

        return render_template('scrapydweb/listspiders.html', node=self.node, results=results)

    def listversions(self):
        results = []
        for version in self.js['versions']:
            try:
                version_readable = ' (%s)' % datetime.datetime.fromtimestamp(int(version)).isoformat()
            except:
                version_readable = ''

            url_listspiders = url_for('projects', node=self.node, opt='listspiders', project=self.project,
                                      version_spider_job=version)
            url_multinode_delversion = url_for('servers', node=self.node, opt='delversion', project=self.project,
                                               version_job=version)
            url_delversion = url_for('projects', node=self.node, opt='delversion', project=self.project,
                                     version_spider_job=version)
            results.append((version, version_readable, url_listspiders, url_multinode_delversion, url_delversion))

        kwargs = dict(
            node=self.node,
            project=self.project,
            results=results,
            url_multinode_delproject=url_for('servers', node=self.node, opt='delproject', project=self.project),
            url_delproject=url_for('projects', node=self.node, opt='delproject', project=self.project)
        )
        return render_template('scrapydweb/listversions.html', **kwargs)



# coding: utf-8
import os
import re

from scrapydweb.vars import APSCHEDULER_DATABASE_URI, DATA_PATH, DATABASE_PATH, ROOT_DIR


def test_option_data_path(app):
    data_path = os.environ.get('DATA_PATH', '')
    if data_path and os.environ.get('TEST_ON_CIRCLECI', 'False').lower() == 'true':
        assert not os.path.isdir(os.path.join(ROOT_DIR, 'data', 'database'))
    assert os.path.isdir(os.path.join(data_path or DATA_PATH, 'database'))


def test_option_database_url(app):
    database_url = os.environ.get('DATABASE_URL', 'sqlite:///' + DATABASE_PATH)
    assert APSCHEDULER_DATABASE_URI.startswith(database_url)
    assert app.config['SQLALCHEMY_DATABASE_URI'].startswith(database_url)
    for value in app.config['SQLALCHEMY_BINDS'].values():
        assert value.startswith(database_url)

    m = re.match(r'sqlite:///(.+)$', database_url)
    assert os.path.isdir(m.group(1) if m else DATABASE_PATH)



# coding: utf-8
from datetime import datetime
import glob
import io
import os
from pprint import pformat
import re
from shutil import copyfile, copyfileobj, rmtree
from subprocess import CalledProcessError
import tarfile
import tempfile
import time
import zipfile

from flask import flash, redirect, render_template, request, url_for
from six import text_type
from six.moves.configparser import Error as ScrapyCfgParseError
from werkzeug.utils import secure_filename

from ...vars import PY2
from ..baseview import BaseView
from .scrapyd_deploy import _build_egg, get_config
from .utils import mkdir_p, slot


SCRAPY_CFG = """
[settings]
default = projectname.settings

[deploy]
url = http://localhost:6800/
project = projectname

"""
folder_project_dict = {}


class DeployView(BaseView):

    def __init__(self):
        super(DeployView, self).__init__()

        self.url = 'http://{}/{}.json'.format(self.SCRAPYD_SERVER, 'addversion')
        self.template = 'scrapydweb/deploy.html'

        self.scrapy_cfg_list = []
        self.project_paths = []
        self.folders = []
        self.projects = []
        self.modification_times = []
        self.latest_folder = ''

    def dispatch_request(self, **kwargs):
        self.set_scrapy_cfg_list()
        self.project_paths = [os.path.dirname(i) for i in self.scrapy_cfg_list]
        self.folders = [os.path.basename(i) for i in self.project_paths]
        self.get_modification_times()
        self.parse_scrapy_cfg()

        kwargs = dict(
            node=self.node,
            url=self.url,
            url_projects=url_for('projects', node=self.node),
            selected_nodes=self.get_selected_nodes(),
            folders=self.folders,
            projects=self.projects,
            modification_times=self.modification_times,
            latest_folder=self.latest_folder,
            SCRAPY_PROJECTS_DIR=self.SCRAPY_PROJECTS_DIR.replace('\\', '/'),
            url_servers=url_for('servers', node=self.node, opt='deploy'),
            url_deploy_upload=url_for('deploy.upload', node=self.node)
        )
        return render_template(self.template, **kwargs)

    def set_scrapy_cfg_list(self):
        # Python 'ascii' codec can't decode byte
        try:
            self.scrapy_cfg_list = glob.glob(os.path.join(self.SCRAPY_PROJECTS_DIR, '*', u'scrapy.cfg'))
        except UnicodeDecodeError:
            if PY2:
                for name in os.listdir(os.path.join(self.SCRAPY_PROJECTS_DIR, u'')):
                    if not isinstance(name, text_type):
                        msg = "Ignore non-unicode filename %s in %s" % (repr(name), self.SCRAPY_PROJECTS_DIR)
                        self.logger.error(msg)
                        flash(msg, self.WARN)
                    else:
                        scrapy_cfg = os.path.join(self.SCRAPY_PROJECTS_DIR, name, u'scrapy.cfg')
                        if os.path.exists(scrapy_cfg):
                            self.scrapy_cfg_list.append(scrapy_cfg)
            else:
                raise
        # '/home/username/Downloads/scrapydweb/scrapydweb/data/demo_projects/\udc8b\udc8billegal/scrapy.cfg'
        # UnicodeEncodeError: 'utf-8' codec can't encode characters in position 64-65: surrogates not allowed
        new_scrapy_cfg_list = []
        for scrapy_cfg in self.scrapy_cfg_list:
            try:
                scrapy_cfg.encode('utf-8')
            except UnicodeEncodeError:
                msg = "Ignore scrapy.cfg in illegal pathname %s" % repr(os.path.dirname(scrapy_cfg))
                self.logger.error(msg)
                flash(msg, self.WARN)
            else:
                new_scrapy_cfg_list.append(scrapy_cfg)
        self.scrapy_cfg_list = new_scrapy_cfg_list

        self.scrapy_cfg_list.sort(key=lambda x: x.lower())

    def get_modification_times(self):
        timestamps = [self.get_modification_time(path) for path in self.project_paths]
        self.modification_times = [datetime.fromtimestamp(ts).strftime('%Y-%m-%dT%H_%M_%S') for ts in timestamps]

        if timestamps:
            max_timestamp_index = timestamps.index(max(timestamps))
            self.latest_folder = self.folders[max_timestamp_index]
            self.logger.debug('latest_folder: %s', self.latest_folder)

    def get_modification_time(self, path, func_walk=os.walk, retry=True):
        # https://stackoverflow.com/a/29685234/10517783
        # https://stackoverflow.com/a/13454267/10517783
        filepath_list = []
        in_top_dir = True
        try:
            for dirpath, dirnames, filenames in func_walk(path):
                if in_top_dir:
                    in_top_dir = False
                    dirnames[:] = [d for d in dirnames if d not in ['build', 'project.egg-info']]
                    filenames = [f for f in filenames
                                 if not (f.endswith('.egg') or f in ['setup.py', 'setup_backup.py'])]
                for filename in filenames:
                    filepath_list.append(os.path.join(dirpath, filename))
        except UnicodeDecodeError:
            msg = "Found illegal filenames in %s" % path
            self.logger.error(msg)
            flash(msg, self.WARN)
            if PY2 and retry:
                return self.get_modification_time(path, func_walk=self.safe_walk, retry=False)
            else:
                raise
        else:
            return max([os.path.getmtime(f) for f in filepath_list] or [time.time()])

    def parse_scrapy_cfg(self):
        for (idx, scrapy_cfg) in enumerate(self.scrapy_cfg_list):
            folder = self.folders[idx]
            key = '%s (%s)' % (folder, self.modification_times[idx])

            project = folder_project_dict.get(key, '')
            if project:
                self.projects.append(project)
                self.logger.debug('Hit %s, project %s', key, project)
                continue
            else:
                project = folder
                try:
                    # lib/configparser.py: def get(self, section, option, *, raw=False, vars=None, fallback=_UNSET):
                    # projectname/scrapy.cfg: [deploy] project = demo
                    # PY2: get() got an unexpected keyword argument 'fallback'
                    # project = get_config(scrapy_cfg).get('deploy', 'project', fallback=folder) or folder
                    project = get_config(scrapy_cfg).get('deploy', 'project')
                except ScrapyCfgParseError as err:
                    self.logger.error("%s parse error: %s", scrapy_cfg, err)
                finally:
                    project = project or folder
                    self.projects.append(project)
                    folder_project_dict[key] = project
                    self.logger.debug('Add %s, project %s', key, project)

        keys_all = list(folder_project_dict.keys())
        keys_exist = ['%s (%s)' % (_folder, _modification_time)
                      for (_folder, _modification_time) in zip(self.folders, self.modification_times)]
        diff = set(keys_all).difference(set(keys_exist))
        for key in diff:
            self.logger.debug('Pop %s, project %s', key, folder_project_dict.pop(key))
        self.logger.debug(self.json_dumps(folder_project_dict))
        self.logger.debug('folder_project_dict length: %s', len(folder_project_dict))


class DeployUploadView(BaseView):
    methods = ['POST']

    def __init__(self):
        super(DeployUploadView, self).__init__()

        self.url = ''
        self.template = 'scrapydweb/deploy_results.html'

        self.folder = ''
        self.project = ''
        self.version = ''
        self.selected_nodes_amount = 0
        self.selected_nodes = []
        self.first_selected_node = 0

        self.eggname = ''
        self.eggpath = ''
        self.scrapy_cfg_path = ''
        self.scrapy_cfg_searched_paths = []
        self.scrapy_cfg_not_found = False
        self.scrapy_cfg_parse_error = ''
        self.build_egg_subprocess_error = ''
        self.data = None
        self.js = {}

        self.slot = slot

    def dispatch_request(self, **kwargs):
        self.handle_form()

        if self.scrapy_cfg_not_found or self.scrapy_cfg_parse_error or self.build_egg_subprocess_error:
            if self.selected_nodes_amount > 1:
                alert = "Multinode deployment terminated:"
            else:
                alert = "Fail to deploy project:"

            if self.scrapy_cfg_not_found:
                text = "scrapy.cfg not found"
                tip = "Make sure that the 'scrapy.cfg' file resides in your project directory. "
            elif self.scrapy_cfg_parse_error:
                text = self.scrapy_cfg_parse_error
                tip = "Check the content of the 'scrapy.cfg' file in your project directory. "
            else:
                text = self.build_egg_subprocess_error
                tip = ("Check the content of the 'scrapy.cfg' file in your project directory. "
                       "Or build the egg file by yourself instead. ")

            if self.scrapy_cfg_not_found:
                # Handle case when scrapy.cfg not found in zip file which contains illegal pathnames in PY3
                message = "scrapy_cfg_searched_paths:\n%s" % pformat(self.scrapy_cfg_searched_paths)
            else:
                message = "# The 'scrapy.cfg' file in your project directory should be like:\n%s" % SCRAPY_CFG

            return render_template(self.template_fail, node=self.node,
                                   alert=alert, text=text, tip=tip, message=message)
        else:
            self.prepare_data()
            status_code, self.js = self.make_request(self.url, data=self.data, auth=self.AUTH)

        if self.js['status'] != self.OK:
            # With multinodes, would try to deploy to the first selected node first
            if self.selected_nodes_amount > 1:
                alert = ("Multinode deployment terminated, "
                         "since the first selected node returned status: " + self.js['status'])
            else:
                alert = "Fail to deploy project, got status: " + self.js['status']
            message = self.js.get('message', '')
            if message:
                self.js['message'] = 'See details below'

            return render_template(self.template_fail, node=self.node,
                                   alert=alert, text=self.json_dumps(self.js), message=message)
        else:
            if self.selected_nodes_amount == 0:
                return redirect(url_for('schedule', node=self.node,
                                        project=self.project, version=self.version))
            else:
                kwargs = dict(
                    node=self.node,
                    selected_nodes=self.selected_nodes,
                    first_selected_node=self.first_selected_node,
                    js=self.js,
                    project=self.project,
                    version=self.version,
                    url_projects_first_selected_node=url_for('projects', node=self.first_selected_node),
                    url_projects_list=[url_for('projects', node=n) for n in range(1, self.SCRAPYD_SERVERS_AMOUNT + 1)],
                    url_xhr=url_for('deploy.xhr', node=self.node, eggname=self.eggname,
                                    project=self.project, version=self.version),
                    url_schedule=url_for('schedule', node=self.node, project=self.project,
                                         version=self.version),
                    url_servers=url_for('servers', node=self.node, opt='schedule', project=self.project,
                                        version_job=self.version)
                )
                return render_template(self.template, **kwargs)

    def handle_form(self):
        # {'1': 'on',
        # '2': 'on',
        # 'checked_amount': '2',
        # 'folder': 'ScrapydWeb_demo',
        # 'project': 'demo',
        # 'version': '2018-09-05T03_13_50'}

        # With multinodes, would try to deploy to the first selected node first
        self.selected_nodes_amount = request.form.get('checked_amount', default=0, type=int)
        if self.selected_nodes_amount:
            self.selected_nodes = self.get_selected_nodes()
            self.first_selected_node = self.selected_nodes[0]
            self.url = 'http://{}/{}.json'.format(self.SCRAPYD_SERVERS[self.first_selected_node - 1], 'addversion')
            # Note that self.first_selected_node != self.node
            self.AUTH = self.SCRAPYD_SERVERS_AUTHS[self.first_selected_node - 1]
        else:
            self.url = 'http://{}/{}.json'.format(self.SCRAPYD_SERVER, 'addversion')

        # Error: Project names must begin with a letter and contain only letters, numbers and underscores
        self.project = re.sub(self.STRICT_NAME_PATTERN, '_', request.form.get('project', '')) or self.get_now_string()
        self.version = re.sub(self.LEGAL_NAME_PATTERN, '-', request.form.get('version', '')) or self.get_now_string()

        if request.files.get('file'):
            self.handle_uploaded_file()
        else:
            self.folder = request.form['folder']  # Used with SCRAPY_PROJECTS_DIR to get project_path
            self.handle_local_project()

    def handle_local_project(self):
        # Use folder instead of project
        project_path = os.path.join(self.SCRAPY_PROJECTS_DIR, self.folder)

        self.search_scrapy_cfg_path(project_path)
        if not self.scrapy_cfg_path:
            self.scrapy_cfg_not_found = True
            return

        self.eggname = '%s_%s.egg' % (self.project, self.version)
        self.eggpath = os.path.join(self.DEPLOY_PATH, self.eggname)
        self.build_egg()

    def handle_uploaded_file(self):
        # http://flask.pocoo.org/docs/1.0/api/#flask.Request.form
        # <class 'werkzeug.datastructures.FileStorage'>
        file = request.files['file']

        # Non-ASCII would be omitted and resulting the filename as to 'egg' or 'tar.gz'
        filename = secure_filename(file.filename)
        # tar.xz only works on Linux and macOS
        if filename in ['egg', 'zip', 'tar.gz']:
            filename = '%s_%s.%s' % (self.project, self.version, filename)
        else:
            filename = '%s_%s_from_file_%s' % (self.project, self.version, filename)

        if filename.endswith('egg'):
            self.eggname = filename
            self.eggpath = os.path.join(self.DEPLOY_PATH, self.eggname)
            file.save(self.eggpath)
            self.scrapy_cfg_not_found = False
        else:  # Compressed file
            filepath = os.path.join(self.DEPLOY_PATH, filename)
            file.save(filepath)
            tmpdir = self.uncompress_to_tmpdir(filepath)

            # Search from the root of tmpdir
            self.search_scrapy_cfg_path(tmpdir)
            if not self.scrapy_cfg_path:
                self.scrapy_cfg_not_found = True
                return

            self.eggname = re.sub(r'(\.zip|\.tar\.gz)$', '.egg', filename)
            self.eggpath = os.path.join(self.DEPLOY_PATH, self.eggname)
            self.build_egg()

    # https://gangmax.me/blog/2011/09/17/12-14-52-publish-532/
    # https://stackoverflow.com/a/49649784
    # When ScrapydWeb runs in Linux/macOS and tries to uncompress zip file from Windows_CN_cp936
    # UnicodeEncodeError: 'ascii' codec can't encode characters in position 7-8: ordinal not in range(128)
    # macOS + PY2 would raise OSError: Illegal byte sequence
    # Ubuntu + PY2 would raise UnicodeDecodeError in search_scrapy_cfg_path() though f.extractall(tmpdir) works well
    def uncompress_to_tmpdir(self, filepath):
        self.logger.debug("Uncompressing %s", filepath)
        tmpdir = tempfile.mkdtemp(prefix="scrapydweb-uncompress-")
        if zipfile.is_zipfile(filepath):
            with zipfile.ZipFile(filepath, 'r') as f:
                if PY2:
                    tmpdir = tempfile.mkdtemp(prefix="scrapydweb-uncompress-")
                    for filename in f.namelist():
                        try:
                            filename_utf8 = filename.decode('gbk').encode('utf8')
                        except (UnicodeDecodeError, UnicodeEncodeError):
                            filename_utf8 = filename
                        filepath_utf8 = os.path.join(tmpdir, filename_utf8)

                        try:
                            with io.open(filepath_utf8, 'wb') as f_utf8:
                                copyfileobj(f.open(filename), f_utf8)
                        except IOError:
                            # os.mkdir(filepath_utf8)
                            # zipfile from Windows "send to zipped" would meet the inner folder first:
                            # temp\\scrapydweb-uncompress-qrcyc0\\demo7/demo/'
                            mkdir_p(filepath_utf8)
                else:
                    f.extractall(tmpdir)
        else:  # tar.gz
            with tarfile.open(filepath, 'r') as tar:  # Open for reading with transparent compression (recommended).
                tar.extractall(tmpdir)

        self.logger.debug("Uncompressed to %s", tmpdir)
        # In case uploading a compressed file in which scrapy_cfg_dir contains none ascii in python 2,
        # whereas selecting a project for auto packaging, scrapy_cfg_dir is unicode
        # print(repr(tmpdir))
        # print(type(tmpdir))
        return tmpdir.decode('utf8') if PY2 else tmpdir

    def search_scrapy_cfg_path(self, search_path, func_walk=os.walk, retry=True):
        try:
            for dirpath, dirnames, filenames in func_walk(search_path):
                self.scrapy_cfg_searched_paths.append(os.path.abspath(dirpath))
                self.scrapy_cfg_path = os.path.abspath(os.path.join(dirpath, 'scrapy.cfg'))
                if os.path.exists(self.scrapy_cfg_path):
                    self.logger.debug("scrapy_cfg_path: %s", self.scrapy_cfg_path)
                    return
        except UnicodeDecodeError:
            msg = "Found illegal filenames in %s" % search_path
            self.logger.error(msg)
            flash(msg, self.WARN)
            if PY2 and retry:
                self.search_scrapy_cfg_path(search_path, func_walk=self.safe_walk, retry=False)
            else:
                raise
        else:
            self.logger.error("scrapy.cfg not found in: %s", search_path)
            self.scrapy_cfg_path = ''

    def build_egg(self):
        try:
            egg, tmpdir = _build_egg(self.scrapy_cfg_path)
        except ScrapyCfgParseError as err:
            self.logger.error(err)
            self.scrapy_cfg_parse_error = err
            return
        except CalledProcessError as err:
            self.logger.error(err)
            self.build_egg_subprocess_error = err
            return

        scrapy_cfg_dir = os.path.dirname(self.scrapy_cfg_path)
        copyfile(egg, os.path.join(scrapy_cfg_dir, self.eggname))
        copyfile(egg, self.eggpath)
        rmtree(tmpdir)
        self.logger.debug("Egg file saved to: %s", self.eggpath)

    def prepare_data(self):
        with io.open(self.eggpath, 'rb') as f:
            content = f.read()
            self.data = {
                'project': self.project,
                'version': self.version,
                'egg': content
            }

        self.slot.add_egg(self.eggname, content)


class DeployXhrView(BaseView):

    def __init__(self):
        super(DeployXhrView, self).__init__()

        self.eggname = self.view_args['eggname']
        self.project = self.view_args['project']
        self.version = self.view_args['version']

        self.url = 'http://{}/{}.json'.format(self.SCRAPYD_SERVER, 'addversion')

        self.slot = slot

    def dispatch_request(self, **kwargs):
        content = self.slot.egg.get(self.eggname)
        # content = None  # For test only
        if not content:
            eggpath = os.path.join(self.DEPLOY_PATH, self.eggname)
            with io.open(eggpath, 'rb') as f:
                content = f.read()

        data = {
            'project': self.project,
            'version': self.version,
            'egg': content
        }
        status_code, js = self.make_request(self.url, data=data, auth=self.AUTH)
        return self.json_dumps(js, as_response=True)



# coding: utf-8
from flask import url_for

from tests.utils import cst, req, switch_scrapyd


# Location: http://127.0.0.1:5000/1/jobs/?ui=mobile
def test_index(app, client):
    with app.test_request_context():
        for __, headers in cst.HEADERS_DICT.items():
            req(app, client, view='index', kws=dict(ui='mobile'), headers=headers,
                location=url_for('jobs', node=1, ui='mobile'))

        for key in ['Chrome', 'iPad']:
            req(app, client, view='index', kws={}, headers=cst.HEADERS_DICT[key],
                location=url_for('servers', node=1))

        for key in ['iPhone', 'Android']:
            req(app, client, view='index', kws={}, headers=cst.HEADERS_DICT[key],
                location=url_for('jobs', node=1, ui='mobile'))


def test_check_browser(app, client):
    req(app, client, view='servers', kws=dict(node=2), headers=cst.HEADERS_DICT['IE'], ins='checkBrowser();')
    req(app, client, view='servers', kws=dict(node=2), headers=cst.HEADERS_DICT['EDGE'], ins='checkBrowser();')


def test_dropdown_for_mobile_device(app, client):
    req(app, client, view='servers', kws=dict(node=2), headers=cst.HEADERS_DICT['Chrome'],
        ins='dropdown.css', nos=['dropdown_mobileui.css', 'handleDropdown();'])
    req(app, client, view='servers', kws=dict(node=2), headers=cst.HEADERS_DICT['iPhone'],
        nos='dropdown.css', ins=['dropdown_mobileui.css', 'handleDropdown();'])
    req(app, client, view='servers', kws=dict(node=2), headers=cst.HEADERS_DICT['iPad'],
        nos='dropdown.css', ins=['dropdown_mobileui.css', 'handleDropdown();'])


def test_check_update(app, client):
    @app.context_processor
    def inject_variable():
        return dict(CHECK_LATEST_VERSION_FREQ=1)

    req(app, client, view='servers', kws=dict(node=2),
        ins='<script>setTimeout("checkLatestVersion(', nos='<!-- <script>setTimeout("checkLatestVersion(')

    @app.context_processor
    def inject_variable():
        return dict(CHECK_LATEST_VERSION_FREQ=100)

    req(app, client, view='servers', kws=dict(node=2), nos='<script>setTimeout("checkLatestVersion(')


def test_page(app, client):
    for view, title in cst.VIEW_TITLE_MAP.items():
        req(app, client, view=view, kws=dict(node=1), ins=title)

    # When v*p*n is globally enabled for testing Telegram, got 500 Internal Privoxy Error
    ins = ['fail - ScrapydWeb', 'status_code: -1']
    for view in ['jobs', 'logs', 'items']:
        req(app, client, view=view, kws=dict(node=2), ins=ins)

    switch_scrapyd(app)

    for view, title in cst.VIEW_TITLE_MAP.items():
        req(app, client, view=view, kws=dict(node=2), ins=title)


def test_select_multinode_checkbox(app, client):
    for view in ['deploy', 'schedule']:
        req(app, client, view=view, kws=dict(node=2), ins='CheckAll / UncheckAll')


def test_items(app, client):
    try:
        req(app, client, view='items', kws=dict(node=1), ins='Directory listing for /items/')
    except AssertionError:
        req(app, client, view='items', kws=dict(node=1), ins='No Such Resource')


def test_switch_node_skip(app, client):
    req(app, client, view='servers', kws=dict(node=1),
        ins=['1 / 2', 'onclick="switchNode(1);', 'id="skip_nodes_checkbox"'])
    req(app, client, view='servers', kws=dict(node=2),
        ins=['2 / 2', 'onclick="switchNode(-1);', 'id="skip_nodes_checkbox"'])


# <span>Cluster Reports</span>
# <el-tab-pane label="Get Reports" name="getreports">
def test_cluster_reports_exists(app, client):
    ins = ['<span>Cluster Reports</span>', '<el-tab-pane label="Get Reports"']
    req(app, client, view='servers', kws=dict(node=1), ins=ins)



# coding: utf-8
from collections import OrderedDict
from email.mime.text import MIMEText
import json
import logging
import smtplib
import sys
import time


logger = logging.getLogger('scrapydweb.utils.send_email')  # __name__
logger.setLevel(logging.DEBUG)


# https://stackoverflow.com/a/27515833/10517783 How to send an email with Gmail as provider using Python?
# https://stackoverflow.com/a/26053352/10517783 Python smtplib proxy support
def send_email(**kwargs):
    to_retry = kwargs.get('to_retry', False)
    need_debug = kwargs.get('need_debug', False)

    email_username = kwargs['email_username']
    email_password = kwargs['email_password']
    email_sender = kwargs['email_sender']
    email_recipients = kwargs['email_recipients']
    smtp_server = kwargs['smtp_server']
    smtp_port = kwargs['smtp_port']
    smtp_over_ssl = kwargs['smtp_over_ssl']
    smtp_connection_timeout = kwargs['smtp_connection_timeout']
    subject = kwargs['subject']
    content = kwargs['content']
    # https://stackoverflow.com/questions/6921699/can-i-get-json-to-load-into-an-ordereddict/6921760#6921760
    # data = json.loads('{"foo":1, "bar": 2}', object_pairs_hook=OrderedDict)
    # In log.py : ensure_ascii=True
    # json.loads('abc') -> JSONDecodeError
    try:
        content = json.dumps(json.loads(content, object_pairs_hook=OrderedDict),
                             sort_keys=False, indent=4, ensure_ascii=False)
    except ValueError:
        pass

    msg = MIMEText(u'%s\n%s' % (time.ctime(), content), 'plain', 'utf-8')
    msg['From'] = email_sender
    msg['Subject'] = u'{} {}'.format(time.strftime('%H:%M'), subject)

    server = None
    result = False
    reason = ''
    try:
        if smtp_over_ssl:
            server = smtplib.SMTP_SSL(smtp_server, smtp_port, timeout=smtp_connection_timeout)
        else:
            server = smtplib.SMTP(smtp_server, smtp_port, timeout=smtp_connection_timeout)
            server.ehlo()
            server.starttls()
        if need_debug:
            server.set_debuglevel(1)  # For debug
        server.login(email_username, email_password)
        server.sendmail(email_sender, email_recipients, msg.as_string())
    except Exception as err:
        logger.error("Fail to send email: %s", subject)
        try:
            reason = err.args[-1].decode('utf8')
        except:
            try:
                reason = err.args[-1].decode('gbk')
            except:
                reason = str(err)
        logger.info("Fail reason: %s", reason)
        if to_retry:
            kwargs.update(to_retry=False, need_debug=True)
            logger.debug("Retrying...")
            time.sleep(3)
            return send_email(**kwargs)
    else:
        result = True
        reason = "Sent"
        logger.info("Email sent: %s", subject)
    finally:
        if server is not None:
            try:
                server.quit()
            except:
                pass

    return result, reason


if __name__ == '__main__':
    # To avoid logging twice when importing the send_email function to send email.
    _handler = logging.StreamHandler()
    _formatter = logging.Formatter(fmt="[%(asctime)s] %(levelname)-8s in %(name)s: %(message)s")
    _handler.setFormatter(_formatter)
    logger.addHandler(_handler)

    send_email(**json.loads(sys.argv[1]))



# coding: utf-8
from collections import OrderedDict, defaultdict
from datetime import date, datetime
import io
import json
import os
import re
from subprocess import Popen
import sys
import tarfile
import time

from flask import flash, get_flashed_messages, render_template, request, url_for
from logparser import parse

from ...vars import ROOT_DIR
from ..baseview import BaseView


EMAIL_CONTENT_KEYS = [
    'log_critical_count',
    'log_error_count',
    'log_warning_count',
    'log_redirect_count',
    'log_retry_count',
    'log_ignore_count',
    'crawled_pages',
    'scraped_items'
]
job_data_dict = {}
# job_finished_key_dict would only be updated by poll POST with ?job_finished=True > monitor_alert(),
# used for determining whether to show 'click to refresh' button in the Log and Stats page.
job_finished_key_dict = defaultdict(OrderedDict)
# For /log/report/
job_finished_report_dict = defaultdict(OrderedDict)
REPORT_KEYS_SET = {'from_memory', 'status', 'pages', 'items', 'shutdown_reason', 'finish_reason', 'runtime',
                   'first_log_time', 'latest_log_time', 'log_categories', 'latest_matches'}


# http://flask.pocoo.org/docs/1.0/api/#flask.views.View
# http://flask.pocoo.org/docs/1.0/views/
class LogView(BaseView):

    def __init__(self):
        super(LogView, self).__init__()  # super().__init__()

        self.opt = self.view_args['opt']
        self.project = self.view_args['project']
        self.spider = self.view_args['spider']
        self.job = self.view_args['job']

        self.job_key = '/%s/%s/%s/%s' % (self.node, self.project, self.spider, self.job)

        # Note that self.LOCAL_SCRAPYD_LOGS_DIR may be an empty string
        # Extension like '.log' is excluded here.
        self.url = u'http://{}/logs/{}/{}/{}'.format(self.SCRAPYD_SERVER, self.project, self.spider, self.job)
        self.log_path = os.path.join(self.LOCAL_SCRAPYD_LOGS_DIR, self.project, self.spider, self.job)

        # For Log and Stats buttons in the Logs page: /a.log/?with_ext=True
        self.with_ext = request.args.get('with_ext', None)
        if self.with_ext:
            self.SCRAPYD_LOG_EXTENSIONS = ['']
            job_without_ext = self.get_job_without_ext(self.job)
        else:
            job_without_ext = self.job

        # json file by LogParser
        self.json_path = os.path.join(self.LOCAL_SCRAPYD_LOGS_DIR, self.project, self.spider, job_without_ext+'.json')
        self.json_url = u'http://{}/logs/{}/{}/{}.json'.format(self.SCRAPYD_SERVER, self.project, self.spider,
                                                               job_without_ext)

        self.status_code = 0
        self.text = ''
        if self.opt == 'report':
            self.template = None
        else:
            self.template = 'scrapydweb/%s%s.html' % (self.opt, '_mobileui' if self.USE_MOBILEUI else '')
        self.kwargs = dict(node=self.node, project=self.project, spider=self.spider,
                           job=job_without_ext, url_refresh='', url_jump='')

        # Request that comes from poll POST for finished job and links of finished job in the Jobs page
        # would be attached with the query string '?job_finished=True'
        self.job_finished = request.args.get('job_finished', None)

        self.utf8_realtime = False
        self.stats_realtime = False
        self.stats_logparser = False
        self.report_logparser = False
        if self.opt == 'utf8':
            flash("It's recommended to check out the latest log via: the Stats page >> View log >> Tail", self.WARN)
            self.utf8_realtime = True
        elif self.opt == 'stats':
            self.stats_realtime = True if request.args.get('realtime', None) else False
            self.stats_logparser = not self.stats_realtime
        else:
            self.report_logparser = True
        self.logparser_valid = False
        self.backup_stats_valid = False
        spider_path = self.mkdir_spider_path()
        self.backup_stats_path = os.path.join(spider_path, job_without_ext + '.json')
        self.stats = {}

        # job_data for monitor & alert: ([0] * 8, [False] * 6, False, time.time())
        self.job_stats_previous = []
        self.triggered_list = []
        self.has_been_stopped = False
        self.last_send_timestamp = 0
        self.job_stats = []
        self.job_stats_diff = []
        # For compatibility with Python 2, use OrderedDict() to keep insertion order
        self.email_content_kwargs = OrderedDict()
        self.flag = ''

        self.jobs_to_keep = self.JOBS_FINISHED_JOBS_LIMIT or 200

    def dispatch_request(self, **kwargs):
        if self.report_logparser:
            self.read_stats_for_report()
        # Try to request stats by LogParser to avoid reading/requesting the whole log
        if not self.logparser_valid and (self.stats_logparser or self.report_logparser):
            if self.IS_LOCAL_SCRAPYD_SERVER and self.LOCAL_SCRAPYD_LOGS_DIR:
                self.read_local_stats_by_logparser()
            if not self.logparser_valid:
                self.request_stats_by_logparser()

        if not self.logparser_valid and not self.text:
            # Try to read local logfile
            if self.IS_LOCAL_SCRAPYD_SERVER and self.LOCAL_SCRAPYD_LOGS_DIR:
                self.read_local_scrapy_log()
            # Has to request scrapy logfile
            if not self.text:
                self.request_scrapy_log()
                if self.status_code != 200:
                    if self.stats_logparser or self.report_logparser:
                        self.load_backup_stats()
                    if not self.backup_stats_valid:
                        if not self.report_logparser:
                            kwargs = dict(node=self.node, url=self.url, status_code=self.status_code, text=self.text)
                            return render_template(self.template_fail, **kwargs)
            else:
                self.url += self.SCRAPYD_LOG_EXTENSIONS[0]
        else:
            self.url += self.SCRAPYD_LOG_EXTENSIONS[0]

        if (not self.utf8_realtime
            and not self.logparser_valid
            and self.text
            and self.status_code in [0, 200]):
            self.logger.warning('Parse the whole log')
            self.stats = parse(self.text)
            # Note that the crawler_engine is not available when using parse()
            self.stats.setdefault('crawler_engine', {})
            self.stats.setdefault('status', self.OK)

        if self.report_logparser:
            if self.stats and not self.stats.setdefault('from_memory', False):
                self.simplify_stats_for_report()
                self.keep_stats_for_report()
            get_flashed_messages()
            # 0, -1, 404 load backup
            if self.status_code < 100 or self.stats:
                status_code = 200
            else:
                status_code = self.status_code
            return self.json_dumps(self.stats or dict(status='error'), as_response=True), status_code
        else:
            self.update_kwargs()
            if self.ENABLE_MONITOR and self.POST:  # Only poll.py would make POST request
                self.monitor_alert()
            return render_template(self.template, **self.kwargs)

    def read_local_stats_by_logparser(self):
        self.logger.debug("Try to read local stats by LogParser: %s", self.json_path)
        try:
            with io.open(self.json_path, 'r', encoding='utf-8') as f:
                js = json.loads(f.read())
        except Exception as err:
            self.logger.error("Fail to read local stats from %s: %s", self.json_path, err)
            return
        else:
            if js.get('logparser_version') != self.LOGPARSER_VERSION:
                msg = "Mismatching logparser_version %s in local stats" % js.get('logparser_version')
                self.logger.warning(msg)
                flash(msg, self.WARN)
                return
            self.logparser_valid = True
            self.stats = js
            msg = "Using local stats: LogParser v%s, last updated at %s, %s" % (
                js['logparser_version'], js['last_update_time'], self.handle_slash(self.json_path))
            self.logger.info(msg)
            flash(msg, self.INFO)

    def request_stats_by_logparser(self):
        self.logger.debug("Try to request stats by LogParser: %s", self.json_url)
        # self.make_request() would check the value of key 'status' if as_json=True
        status_code, js = self.make_request(self.json_url, auth=self.AUTH, as_json=True, dumps_json=False)
        if status_code != 200:
            self.logger.error("Fail to request stats from %s, got status_code: %s", self.json_url, status_code)
            if self.IS_LOCAL_SCRAPYD_SERVER and self.ENABLE_LOGPARSER:
                flash("Request to %s got code %s, wait until LogParser parses the log. " % (self.json_url, status_code),
                      self.INFO)
            else:
                flash(("'pip install logparser' on host '%s' and run command 'logparser'. "
                       "Or wait until LogParser parses the log. ") % self.SCRAPYD_SERVER, self.WARN)
            return
        elif js.get('logparser_version') != self.LOGPARSER_VERSION:
            msg = "'pip install --upgrade logparser' on host '%s' to update LogParser to v%s" % (
                self.SCRAPYD_SERVER, self.LOGPARSER_VERSION)
            self.logger.warning(msg)
            flash(msg, self.WARN)
            return
        else:
            self.logparser_valid = True
            # TODO: dirty data
            self.stats = js
            msg = "LogParser v%s, last updated at %s, %s" % (
                js['logparser_version'], js['last_update_time'], self.json_url)
            self.logger.info(msg)
            flash(msg, self.INFO)

    def read_local_scrapy_log(self):
        for ext in self.SCRAPYD_LOG_EXTENSIONS:
            log_path = self.log_path + ext
            if os.path.exists(log_path):
                if tarfile.is_tarfile(log_path):
                    self.logger.debug("Ignore local tarfile and use requests instead: %s", log_path)
                    break
                with io.open(log_path, 'r', encoding='utf-8', errors='ignore') as f:
                    self.text = f.read()
                log_path = self.handle_slash(log_path)
                msg = "Using local logfile: %s" % log_path
                self.logger.debug(msg)
                flash(msg, self.INFO)
                break

    def request_scrapy_log(self):
        for ext in self.SCRAPYD_LOG_EXTENSIONS:
            url = self.url + ext
            self.status_code, self.text = self.make_request(url, auth=self.AUTH, as_json=False)
            if self.status_code == 200:
                self.url = url
                self.logger.debug("Got logfile from %s", self.url)
                break
        else:
            msg = "Fail to request logfile from %s with extensions %s" % (self.url, self.SCRAPYD_LOG_EXTENSIONS)
            self.logger.error(msg)
            flash(msg, self.WARN)
            self.url += self.SCRAPYD_LOG_EXTENSIONS[0]

    def simplify_stats_for_report(self):
        for key in list(self.stats.keys()):
            if key not in REPORT_KEYS_SET:
                self.stats.pop(key)
        try:
            for key in self.stats['log_categories']:
                self.stats['log_categories'][key] = dict(count=self.stats['log_categories'][key]['count'])
        except KeyError:
            pass
        try:
            self.stats['latest_matches'] = dict(latest_item=self.stats['latest_matches']['latest_item'])
        except KeyError:
            pass

    def keep_stats_for_report(self):
        od = job_finished_report_dict[self.node]
        if self.job_key in od:
            return
        if (self.stats.get('shutdown_reason', self.NA) == self.NA
            and self.stats.get('finish_reason', self.NA) == self.NA):
            return
        if set(self.stats.keys()) == REPORT_KEYS_SET:
            od[self.job_key] = self.stats
            if len(od) > self.jobs_to_keep:
                od.popitem(last=False)
            self.logger.debug("%s keys in job_finished_report_dict[%s]", len(od), self.node)

    def read_stats_for_report(self):
        try:
            self.stats = job_finished_report_dict[self.node][self.job_key]
        except KeyError:
            self.logger.debug("%s not found in job_finished_report_dict[%s]", self.job_key, self.node)
        else:
            self.logger.debug("%s found in job_finished_report_dict[%s]", self.job_key, self.node)
            self.logparser_valid = True
            self.stats['from_memory'] = True

    def mkdir_spider_path(self):
        node_path = os.path.join(self.STATS_PATH,
                                 re.sub(self.LEGAL_NAME_PATTERN, '-', re.sub(r'[.:]', '_', self.SCRAPYD_SERVER)))
        project_path = os.path.join(node_path, self.project)
        spider_path = os.path.join(project_path, self.spider)

        if not os.path.isdir(self.STATS_PATH):
            os.mkdir(self.STATS_PATH)
        if not os.path.isdir(node_path):
            os.mkdir(node_path)
        if not os.path.isdir(project_path):
            os.mkdir(project_path)
        if not os.path.isdir(spider_path):
            os.mkdir(spider_path)
        return spider_path

    def backup_stats(self):
        # TODO: delete backup stats json file when the job is deleted in the Jobs page with database view
        try:
            with io.open(self.backup_stats_path, 'w', encoding='utf-8', errors='ignore') as f:
                f.write(self.json_dumps(self.stats))
        except Exception as err:
            self.logger.error("Fail to backup stats to %s: %s" % (self.backup_stats_path, err))
            try:
                os.remove(self.backup_stats_path)
            except:
                pass
        else:
            self.logger.info("Saved backup stats to %s", self.backup_stats_path)

    def load_backup_stats(self):
        self.logger.debug("Try to load backup stats by LogParser: %s", self.json_path)
        try:
            with io.open(self.backup_stats_path, 'r', encoding='utf-8') as f:
                js = json.loads(f.read())
        except Exception as err:
            self.logger.error("Fail to load backup stats from %s: %s", self.backup_stats_path, err)
        else:
            if js.get('logparser_version') != self.LOGPARSER_VERSION:
                msg = "Mismatching logparser_version %s in backup stats" % js.get('logparser_version')
                self.logger.warning(msg)
                flash(msg, self.WARN)
                return
            self.logparser_valid = True
            self.backup_stats_valid = True
            self.stats = js
            msg = "Using backup stats: LogParser v%s, last updated at %s, %s" % (
                js['logparser_version'], js['last_update_time'], self.handle_slash(self.backup_stats_path))
            self.logger.info(msg)
            flash(msg, self.WARN)

    @staticmethod
    def get_ordered_dict(adict):
        # 'source', 'last_update_time', 'last_update_timestamp', other keys in order
        odict = OrderedDict()
        for k in ['source', 'last_update_time', 'last_update_timestamp']:
            odict[k] = adict.pop(k)
        for k in sorted(adict.keys()):
            odict[k] = adict[k]
        return odict

    def update_kwargs(self):
        if self.utf8_realtime:
            self.kwargs['text'] = self.text
            self.kwargs['last_update_timestamp'] = time.time()
            if self.job_finished or self.job_key in job_finished_key_dict[self.node]:
                self.kwargs['url_refresh'] = ''
            else:
                self.kwargs['url_refresh'] = 'javascript:location.reload(true);'
        else:
            # Parsed data comes from json.loads, for compatibility with Python 2,
            # use str(time_) to avoid [u'2019-01-01 00:00:01', 0, 0, 0, 0] in JavaScript.
            for d in self.stats['datas']:
                d[0] = str(d[0])
            # For sorted orders in stats.html with Python 2
            for k in ['crawler_stats', 'crawler_engine']:
                if self.stats[k]:
                    self.stats[k] = self.get_ordered_dict(self.stats[k])

            if self.BACKUP_STATS_JSON_FILE:
                self.backup_stats()
            self.kwargs.update(self.stats)

            if (self.kwargs['finish_reason'] == self.NA
               and not self.job_finished
               and self.job_key not in job_finished_key_dict[self.node]):
                # http://flask.pocoo.org/docs/1.0/api/#flask.Request.url_root
                # _query_string = '?ui=mobile'
                # self.url_refresh = request.script_root + request.path + _query_string
                self.kwargs['url_refresh'] = 'javascript:location.reload(true);'
            if self.kwargs['url_refresh']:
                if self.stats_logparser and not self.logparser_valid:
                    self.kwargs['url_jump'] = ''
                else:
                    self.kwargs['url_jump'] = url_for('log', node=self.node, opt='stats', project=self.project,
                                                      spider=self.spider, job=self.job, with_ext=self.with_ext,
                                                      ui=self.UI, realtime='True' if self.stats_logparser else None)

        # Stats link of 'a.json' from the Logs page should hide these links
        if self.with_ext and self.job.endswith('.json'):
            self.kwargs['url_source'] = ''
            self.kwargs['url_opt_opposite'] = ''
            self.kwargs['url_refresh'] = ''
            self.kwargs['url_jump'] = ''
        else:
            if self.SCRAPYD_SERVER_PUBLIC_URL:
                self.kwargs['url_source'] = re.sub(r'^http.*?/logs/', self.SCRAPYD_SERVER_PUBLIC_URL + '/logs/',
                                                   self.url)
            else:
                self.kwargs['url_source'] = self.url
            self.kwargs['url_opt_opposite'] = url_for('log', node=self.node,
                                                      opt='utf8' if self.opt == 'stats' else 'stats',
                                                      project=self.project, spider=self.spider, job=self.job,
                                                      job_finished=self.job_finished, with_ext=self.with_ext,
                                                      ui=self.UI)

    # TODO: https://blog.miguelgrinberg.com/post/the-flask-mega-tutorial-part-x-email-support
    def monitor_alert(self):
        job_data_default = ([0] * 8, [False] * 6, False, time.time())
        job_data = job_data_dict.setdefault(self.job_key, job_data_default)
        (self.job_stats_previous, self.triggered_list, self.has_been_stopped, self.last_send_timestamp) = job_data
        self.logger.debug(job_data_dict)
        self.job_stats = [self.kwargs['log_categories'][k.lower() + '_logs']['count']
                          for k in self.ALERT_TRIGGER_KEYS]
        self.job_stats.extend([self.kwargs['pages'] or 0, self.kwargs['items'] or 0])  # May be None by LogParser
        self.job_stats_diff = [j - i for i, j in zip(self.job_stats_previous, self.job_stats)]

        self.set_email_content_kwargs()
        self.set_monitor_flag()
        self.send_alert()
        self.handle_data()

    def set_email_content_kwargs(self):
        self.email_content_kwargs['SCRAPYD_SERVER'] = self.SCRAPYD_SERVER
        self.email_content_kwargs['project'] = self.kwargs['project']
        self.email_content_kwargs['spider'] = self.kwargs['spider']
        self.email_content_kwargs['job'] = self.kwargs['job']
        self.email_content_kwargs['first_log_time'] = self.kwargs['first_log_time']
        self.email_content_kwargs['latest_log_time'] = self.kwargs['latest_log_time']
        self.email_content_kwargs['runtime'] = self.kwargs['runtime']
        self.email_content_kwargs['shutdown_reason'] = self.kwargs['shutdown_reason']
        self.email_content_kwargs['finish_reason'] = self.kwargs['finish_reason']
        self.email_content_kwargs['url_stats'] = request.url + '%sui=mobile' % '&' if request.args else '?'

        for idx, key in enumerate(EMAIL_CONTENT_KEYS):
            if self.job_stats_diff[idx]:
                self.email_content_kwargs[key] = '%s + %s' % (self.job_stats_previous[idx], self.job_stats_diff[idx])
            else:
                self.email_content_kwargs[key] = self.job_stats[idx]
        # pages and items may be None by LogParser
        if self.kwargs['pages'] is None:
            self.email_content_kwargs['crawled_pages'] = self.NA
        if self.kwargs['items'] is None:
            self.email_content_kwargs['scraped_items'] = self.NA

        _url_stop = url_for('api', node=self.node, opt='stop', project=self.project, version_spider_job=self.job)
        self.email_content_kwargs['url_stop'] = self.URL_SCRAPYDWEB + _url_stop

        now_timestamp = time.time()
        for k in ['latest_crawl', 'latest_scrape', 'latest_log']:
            ts = self.kwargs['%s_timestamp' % k]
            self.email_content_kwargs[k] = self.NA if ts == 0 else "%s secs ago" % int(now_timestamp - ts)

        self.email_content_kwargs['current_time'] = self.get_now_string(True)
        self.email_content_kwargs['logparser_version'] = self.kwargs['logparser_version']
        self.email_content_kwargs['latest_item'] = self.kwargs['latest_matches']['latest_item'] or self.NA
        self.email_content_kwargs['Crawler.stats'] = self.kwargs['crawler_stats']
        self.email_content_kwargs['Crawler.engine'] = self.kwargs['crawler_engine']

    def set_monitor_flag(self):
        if self.ON_JOB_FINISHED and self.job_finished:
            self.flag = 'Finished'
        elif not all(self.triggered_list):
            to_forcestop = False
            to_stop = False
            # The order of the elements in ALERT_TRIGGER_KEYS matters:
            # ['CRITICAL', 'ERROR', 'WARNING', 'REDIRECT', 'RETRY', 'IGNORE']
            for idx, key in enumerate(self.ALERT_TRIGGER_KEYS):
                if (0 < getattr(self, 'LOG_%s_THRESHOLD' % key, 0) <= self.job_stats[idx]
                   and not self.triggered_list[idx]):
                    self.triggered_list[idx] = True
                    self.email_content_kwargs['log_%s_count' % key.lower()] += ' triggered!!!'
                    if getattr(self, 'LOG_%s_TRIGGER_FORCESTOP' % key):
                        self.flag = '%s_ForceStop' % key if '_ForceStop' not in self.flag else self.flag
                        to_forcestop = True
                    elif getattr(self, 'LOG_%s_TRIGGER_STOP' % key) and not self.has_been_stopped:
                        self.flag = '%s_Stop' % key if 'Stop' not in self.flag else self.flag
                        self.has_been_stopped = True  # Execute 'Stop' one time at most to avoid unclean shutdown
                        to_stop = True
                    elif not self.has_been_stopped:
                        self.flag = '%s_Trigger' % key if not self.flag else self.flag
            if to_forcestop:
                self.logger.debug("%s: %s", self.flag, self.job_key)
                _url = url_for('api', node=self.node, opt='forcestop',
                               project=self.project, version_spider_job=self.job)
                self.get_response_from_view(_url)
            elif to_stop:
                self.logger.debug("%s: %s", self.flag, self.job_key)
                _url = url_for('api', node=self.node, opt='stop',
                               project=self.project, version_spider_job=self.job)
                self.get_response_from_view(_url)

        if not self.flag and 0 < self.ON_JOB_RUNNING_INTERVAL <= time.time() - self.last_send_timestamp:
            self.flag = 'Running'

    def send_alert(self):
        if (self.flag
            and date.isoweekday(date.today()) in self.ALERT_WORKING_DAYS  # date.isoweekday(datetime.now())
            and datetime.now().hour in self.ALERT_WORKING_HOURS
        ):
            kwargs = dict(
                flag=self.flag,
                pages=self.NA if self.kwargs['pages'] is None else self.kwargs['pages'],
                items=self.NA if self.kwargs['items'] is None else self.kwargs['items'],
                job_key=self.job_key,
                latest_item=self.kwargs['latest_matches']['latest_item'][:100] or self.NA
            )
            subject = u"{flag} [{pages}p, {items}i] {job_key} {latest_item} #scrapydweb".format(**kwargs)
            self.EMAIL_KWARGS['subject'] = subject
            self.EMAIL_KWARGS['content'] = self.json_dumps(self.email_content_kwargs, sort_keys=False)

            data = dict(
                subject=subject,
                url_stats=self.email_content_kwargs['url_stats'],
                url_stop=self.email_content_kwargs['url_stop'],
                when=self.get_now_string(True),
            )
            if self.ENABLE_SLACK_ALERT:
                self.logger.info("Sending alert via Slack: %s", subject)
                _url = url_for('sendtextapi', opt='slack', channel_chatid_subject=None, text=None)
                self.get_response_from_view(_url, data=data)
            if self.ENABLE_TELEGRAM_ALERT:
                self.logger.info("Sending alert via Telegram: %s", subject)
                _url = url_for('sendtextapi', opt='telegram', channel_chatid_subject=None, text=None)
                self.get_response_from_view(_url, data=data)
            if self.ENABLE_EMAIL_ALERT:
                self.logger.info("Sending alert via Email: %s", subject)
                args = [
                    sys.executable,
                    os.path.join(ROOT_DIR, 'utils', 'send_email.py'),
                    self.json_dumps(self.EMAIL_KWARGS, ensure_ascii=True)
                ]
                Popen(args)

    def handle_data(self):
        if self.flag:
            # Update job_data_dict (last_send_timestamp would be updated only when flag is non-empty)
            self.logger.debug("Previous job_data['%s'] %s", self.job_key, job_data_dict[self.job_key])
            job_data_dict[self.job_key] = (self.job_stats, self.triggered_list, self.has_been_stopped, time.time())
            self.logger.debug("Updated  job_data['%s'] %s", self.job_key, job_data_dict[self.job_key])

        if self.job_finished:
            job_data_dict.pop(self.job_key)
            od = job_finished_key_dict[self.node]
            od[self.job_key] = None
            if len(od) > self.jobs_to_keep:
                od.popitem(last=False)
            self.logger.info('job_finished: %s', self.job_key)



# coding: utf-8
from tests.utils import req


projects = [
    'demo_inner',
    'demo_only_scrapy_cfg',
    'demo_outer',
    'demo_unicode',
    'demo___Win10cp1252',
    'demo___Win7CNsendzipped',
    'demo_____',
    'demo________macOS',
    'demo________Ubuntu',
    'demo________Win10cp936',
    'demo________Win7CN',
    'demo________Win7CNsendzipped',
    'fakeproject',
    'ScrapydWeb_demo',
    'scrapy_cfg_no_deploy_project',
    'scrapy_cfg_no_option_project',
    'scrapy_cfg_no_option_project_value',
    'scrapy_cfg_no_section_deploy'
]


def test_cleantest(app, client):
    req(app, client, view='api', kws=dict(node=1, opt='listprojects'))

    for project in projects:
        __, js = req(app, client, view='api', kws=dict(node=1, opt='listjobs', project=project))
        for job in js.get('running', []):
            req(app, client, view='api',
                kws=dict(node=1, opt='forcestop', project=project, version_spider_job=job['id']))
        req(app, client, view='api', kws=dict(node=1, opt='delproject', project=project))



# coding: utf-8
import logging
from logging.config import dictConfig
import platform
import re
import time
import traceback

from flask import Flask, current_app, render_template, url_for
from flask_compress import Compress
from logparser import __version__ as LOGPARSER_VERSION

from .__version__ import __url__, __version__
from .common import handle_metadata
from .models import Metadata, db
from .vars import PYTHON_VERSION, SQLALCHEMY_BINDS, SQLALCHEMY_DATABASE_URI
# from .utils.scheduler import scheduler


# https://stackoverflow.com/questions/18820274/how-to-suppress-sqlalchemy-engine-base-engine-logging-to-stdout
# logging.getLogger('sqlalchemy.engine.base.Engine').propagate = False
logging.getLogger('sqlalchemy.engine.base.Engine').setLevel(logging.WARNING)
# http://flask.pocoo.org/docs/1.0/logging/#basic-configuration
dictConfig({
    'version': 1,
    'formatters': {'default': {
        'format': '[%(asctime)s] %(levelname)-8s in %(name)s: %(message)s',
    }},
    'handlers': {'wsgi': {
        'class': 'logging.StreamHandler',
        'stream': 'ext://flask.logging.wsgi_errors_stream',
        'formatter': 'default'
    }},
    'root': {
        'level': 'DEBUG',
        'handlers': ['wsgi']
    }
})

# Comment out the dictConfig above first
# https://docs.sqlalchemy.org/en/latest/core/engines.html#configuring-logging
# https://apscheduler.readthedocs.io/en/latest/userguide.html#troubleshooting
# logging.basicConfig()
# logging.getLogger('apscheduler').setLevel(logging.DEBUG)
# logging.getLogger('sqlalchemy.engine').setLevel(logging.DEBUG)


def internal_server_error(error):
    kwargs = dict(
        error=error,
        traceback=traceback.format_exc(),
        url_issues=__url__ + '/issues',
        os=platform.platform(),
        python_version=PYTHON_VERSION,
        scrapydweb_version=__version__,
        logparser_version=LOGPARSER_VERSION,
        scrapyd_servers_amount=len(current_app.config.get('SCRAPYD_SERVERS', []))
    )
    return render_template('500.html', **kwargs), 500


def create_app(test_config=None):
    app = Flask(__name__, instance_relative_config=True)
    app.config.from_mapping(
        SECRET_KEY='dev',
    )
    # http://flask.pocoo.org/docs/1.0/config/#configuring-from-files
    app.config.from_object('scrapydweb.default_settings')

    if test_config is None:
        # load the instance config, if it exists, when not testing
        app.config.from_pyfile('config.py', silent=True)
    else:
        # load the test config if passed in
        app.config.from_mapping(test_config)

    @app.route('/hello')
    def hello():
        return 'Hello, World!'

    handle_db(app)
    handle_route(app)
    handle_template_context(app)

    # @app.errorhandler(404)
    # def handle_error(error):
        # return ('Nothing Found', 404)
    # http://flask.pocoo.org/docs/1.0/patterns/errorpages/
    app.register_error_handler(500, internal_server_error)

    # https://ansible-docs.readthedocs.io/zh/stable-2.0/rst/playbooks_filters.html#other-useful-filters
    # https://stackoverflow.com/questions/12791216/how-do-i-use-regular-expressions-in-jinja2
    # https://www.michaelcho.me/article/custom-jinja-template-filters-in-flask
    # http://flask.pocoo.org/docs/1.0/api/#flask.Flask.template_filter
    @app.template_filter()
    def regex_replace(s, find, replace):
        return re.sub(find, replace, s)
    app.jinja_env.variable_start_string = '{{ '
    app.jinja_env.variable_end_string = ' }}'

    compress = Compress()
    compress.init_app(app)

    app.logger.setLevel(logging.DEBUG)
    return app


def handle_db(app):
    # https://flask-sqlalchemy.palletsprojects.com/en/master/config/
    app.config['SQLALCHEMY_DATABASE_URI'] = SQLALCHEMY_DATABASE_URI
    app.config['SQLALCHEMY_BINDS'] = SQLALCHEMY_BINDS
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False  # https://stackoverflow.com/a/33790196/10517783
    app.config['SQLALCHEMY_ECHO'] = True  # http://flask-sqlalchemy.pocoo.org/2.3/config/

    # flask_sqlalchemy/__init__.py
    # class SQLAlchemy(object):
    #     def __init__(self, app=None
    #         self.app = app
    #         if app is not None:
    #             self.init_app(app)
    db.app = app  # https://github.com/viniciuschiele/flask-apscheduler/blob/master/examples/flask_context.py
    db.init_app(app)  # http://flask-sqlalchemy.pocoo.org/2.3/contexts/
    db.create_all()

    # https://blog.miguelgrinberg.com/post/the-flask-mega-tutorial-part-vii-error-handling
    @app.teardown_request
    def handle_db_session(exception):
        if exception:
            db.session.rollback()
        db.session.remove()

    with db.app.app_context():
        if not Metadata.query.filter_by(version=__version__).first():
            metadata = Metadata(version=__version__)
            db.session.add(metadata)
            db.session.commit()
    if time.time() - handle_metadata().get('last_check_update_timestamp', time.time()) > 3600 * 24 * 30:
        handle_metadata('last_check_update_timestamp', time.time())
        handle_metadata('pageview', 0)
    else:
        handle_metadata('pageview', 1)
    # print(Metadata.query.filter_by(version=__version__).first())


def handle_route(app):
    def register_view(view, endpoint, url_defaults_list, with_node=True, trailing_slash=True):
        view_func = view.as_view(endpoint)
        for url, defaults in url_defaults_list:
            rule = '/<int:node>/%s' % url if with_node else '/%s' % url
            if trailing_slash:
                rule += '/'
            if not with_node:
                if defaults:
                    defaults['node'] = 1
                else:
                    defaults = dict(node=1)
            app.add_url_rule(rule, defaults=defaults, view_func=view_func)

    from .views.index import IndexView
    index_view = IndexView.as_view('index')
    app.add_url_rule('/<int:node>/', view_func=index_view)
    app.add_url_rule('/', defaults=dict(node=1), view_func=index_view)

    from .views.api import ApiView
    register_view(ApiView, 'api', [
        ('api/<opt>/<project>/<version_spider_job>', None),
        ('api/<opt>/<project>', dict(version_spider_job=None)),
        ('api/<opt>', dict(project=None, version_spider_job=None))
    ])

    from .views.baseview import MetadataView
    register_view(MetadataView, 'metadata', [('metadata', None)])

    # Overview
    from .views.overview.servers import ServersView
    register_view(ServersView, 'servers', [
        ('servers/getreports/<project>/<spider>/<version_job>', dict(opt='getreports')),
        ('servers/<opt>/<project>/<version_job>/<spider>', None),
        ('servers/<opt>/<project>/<version_job>', dict(spider=None)),
        ('servers/<opt>/<project>', dict(version_job=None, spider=None)),
        ('servers/<opt>', dict(project=None, version_job=None, spider=None)),
        ('servers', dict(opt=None, project=None, version_job=None, spider=None))
    ])

    from .views.overview.multinode import MultinodeView
    register_view(MultinodeView, 'multinode', [
        ('multinode/<opt>/<project>/<version_job>', None),
        ('multinode/<opt>/<project>', dict(version_job=None))
    ])

    from .views.overview.tasks import TasksView, TasksXhrView
    register_view(TasksView, 'tasks', [
        ('tasks/<int:task_id>/<int:task_result_id>', None),
        ('tasks/<int:task_id>', dict(task_result_id=None)),
        ('tasks', dict(task_id=None, task_result_id=None))
    ])
    register_view(TasksXhrView, 'tasks.xhr', [
        ('tasks/xhr/<action>/<int:task_id>/<int:task_result_id>', None),
        ('tasks/xhr/<action>/<int:task_id>', dict(task_result_id=None)),
        ('tasks/xhr/<action>', dict(task_id=None, task_result_id=None))
    ])

    from .views.overview.tasks import bp as bp_tasks_history
    app.register_blueprint(bp_tasks_history)

    # Dashboard
    from .views.dashboard.jobs import JobsView, JobsXhrView
    register_view(JobsView, 'jobs', [('jobs', None)])
    register_view(JobsXhrView, 'jobs.xhr', [('jobs/xhr/<action>/<int:id>', None)])

    from .views.dashboard.node_reports import NodeReportsView
    register_view(NodeReportsView, 'nodereports', [('nodereports', None)])

    from .views.dashboard.cluster_reports import ClusterReportsView
    register_view(ClusterReportsView, 'clusterreports', [
        ('clusterreports/<project>/<spider>/<job>', None),
        ('clusterreports', dict(project=None, spider=None, job=None))
    ])

    # Operations
    from .views.operations.deploy import DeployView, DeployUploadView, DeployXhrView
    register_view(DeployView, 'deploy', [('deploy', None)])
    register_view(DeployUploadView, 'deploy.upload', [('deploy/upload', None)])
    register_view(DeployXhrView, 'deploy.xhr', [('deploy/xhr/<eggname>/<project>/<version>', None)])

    from .views.operations.schedule import (ScheduleView, ScheduleCheckView, ScheduleRunView,
                                            ScheduleXhrView, ScheduleTaskView)
    register_view(ScheduleView, 'schedule', [
        ('schedule/<project>/<version>/<spider>', None),
        ('schedule/<project>/<version>', dict(spider=None)),
        ('schedule/<project>', dict(version=None, spider=None)),
        ('schedule', dict(project=None, version=None, spider=None))
    ])
    register_view(ScheduleCheckView, 'schedule.check', [('schedule/check', None)])
    register_view(ScheduleRunView, 'schedule.run', [('schedule/run', None)])
    register_view(ScheduleXhrView, 'schedule.xhr', [('schedule/xhr/<filename>', None)])
    register_view(ScheduleTaskView, 'schedule.task', [('schedule/task', None)])

    from .views.operations.schedule import bp as bp_schedule_history
    app.register_blueprint(bp_schedule_history)

    # Files
    from .views.files.log import LogView
    register_view(LogView, 'log', [('log/<opt>/<project>/<spider>/<job>', None)])

    from .views.files.logs import LogsView
    register_view(LogsView, 'logs', [
        ('logs/<project>/<spider>', None),
        ('logs/<project>', dict(spider=None)),
        ('logs', dict(project=None, spider=None))
    ])

    from .views.files.items import ItemsView
    register_view(ItemsView, 'items', [
        ('items/<project>/<spider>', None),
        ('items/<project>', dict(spider=None)),
        ('items', dict(project=None, spider=None))
    ])

    from .views.files.projects import ProjectsView
    register_view(ProjectsView, 'projects', [
        ('projects/<opt>/<project>/<version_spider_job>', None),
        ('projects/<opt>/<project>', dict(version_spider_job=None)),
        ('projects', dict(opt='listprojects', project=None, version_spider_job=None))
    ])

    # Parse Log
    from .views.utilities.parse import UploadLogView, UploadedLogView
    register_view(UploadLogView, 'parse.upload', [('parse/upload', None)])
    register_view(UploadedLogView, 'parse.uploaded', [('parse/uploaded/<filename>', None)])

    from .views.utilities.parse import bp as bp_parse_source
    app.register_blueprint(bp_parse_source)

    # Send text
    from .views.utilities.send_text import SendTextView, SendTextApiView
    register_view(SendTextView, 'sendtext', [('sendtext', None)])
    register_view(SendTextApiView, 'sendtextapi', [
        ('slack/<channel_chatid_subject>/<text>', dict(opt='slack')),
        ('slack/<text>', dict(opt='slack', channel_chatid_subject=None)),
        ('slack', dict(opt='slack', channel_chatid_subject=None, text=None)),
        ('telegram/<channel_chatid_subject>/<text>', dict(opt='telegram')),
        ('telegram/<text>', dict(opt='telegram', channel_chatid_subject=None)),
        ('telegram', dict(opt='telegram', channel_chatid_subject=None, text=None)),
        ('tg/<channel_chatid_subject>/<text>', dict(opt='tg')),
        ('tg/<text>', dict(opt='tg', channel_chatid_subject=None)),
        ('tg', dict(opt='tg', channel_chatid_subject=None, text=None)),
        ('email/<channel_chatid_subject>/<text>', dict(opt='email')),
        ('email/<text>', dict(opt='email', channel_chatid_subject=None)),
        ('email', dict(opt='email', channel_chatid_subject=None, text=None)),
    ], with_node=False, trailing_slash=False)

    # System
    from .views.system.settings import SettingsView
    register_view(SettingsView, 'settings', [('settings', None)])


def handle_template_context(app):
    STATIC = 'static'
    VERSION = 'v' + __version__.replace('.', '')
    # MUST be commented out for released version
    # VERSION = 'v131dev'

    @app.context_processor
    def inject_variable():
        return dict(
            CHECK_LATEST_VERSION_FREQ=100,
            GITHUB_URL=__url__,
            PYTHON_VERSION=PYTHON_VERSION,
            SCRAPYDWEB_VERSION=__version__,

            # static_css_common=url_for(STATIC, filename='%s/css/common.css' % VERSION),
            static_css_dropdown=url_for(STATIC, filename='%s/css/dropdown.css' % VERSION),
            static_css_dropdown_mobileui=url_for(STATIC, filename='%s/css/dropdown_mobileui.css' % VERSION),
            static_css_icon_upload_icon_right=url_for(STATIC,
                                                      filename='%s/css/icon_upload_icon_right.css' % VERSION),
            static_css_multinode=url_for(STATIC, filename='%s/css/multinode.css' % VERSION),
            static_css_stacktable=url_for(STATIC, filename='%s/css/stacktable.css' % VERSION),
            static_css_stats=url_for(STATIC, filename='%s/css/stats.css' % VERSION),
            static_css_style=url_for(STATIC, filename='%s/css/style.css' % VERSION),
            static_css_style_mobileui=url_for(STATIC, filename='%s/css/style_mobileui.css' % VERSION),
            static_css_utf8=url_for(STATIC, filename='%s/css/utf8.css' % VERSION),
            static_css_utf8_mobileui=url_for(STATIC, filename='%s/css/utf8_mobileui.css' % VERSION),

            static_css_element_ui_index=url_for(STATIC,
                                                filename='%s/element-ui@2.4.6/lib/theme-chalk/index.css' % VERSION),
            static_js_element_ui_index=url_for(STATIC, filename='%s/element-ui@2.4.6/lib/index.js' % VERSION),

            static_js_common=url_for(STATIC, filename='%s/js/common.js' % VERSION),
            static_js_echarts_min=url_for(STATIC, filename='%s/js/echarts.min.js' % VERSION),
            static_js_icons_menu=url_for(STATIC, filename='%s/js/icons_menu.js' % VERSION),
            # static_js_github_buttons_html=url_for(STATIC, filename='%s/js/github_buttons.html' % VERSION),
            static_js_github_buttons=url_for(STATIC, filename='%s/js/github_buttons.js' % VERSION),
            static_js_jquery_min=url_for(STATIC, filename='%s/js/jquery.min.js' % VERSION),
            static_js_multinode=url_for(STATIC, filename='%s/js/multinode.js' % VERSION),
            static_js_stacktable=url_for(STATIC, filename='%s/js/stacktable.js' % VERSION),
            static_js_stats=url_for(STATIC, filename='%s/js/stats.js' % VERSION),
            static_js_vue_min=url_for(STATIC, filename='%s/js/vue.min.js' % VERSION),

            static_icon=url_for(STATIC, filename='%s/icon/fav.ico' % VERSION),
            static_icon_shortcut=url_for(STATIC, filename='%s/icon/fav.ico' % VERSION),
            static_icon_apple_touch=url_for(STATIC, filename='%s/icon/spiderman.png' % VERSION),
        )



# coding: utf-8
from datetime import datetime, timedelta
import re

from flask import url_for
from six.moves.urllib.parse import unquote_plus

from tests.utils import cst, req, req_single_scrapyd, sleep, switch_scrapyd, upload_file_deploy


NODE = 2
metadata = {}
FILENAME = '%s_%s_%s.pickle' % (cst.PROJECT, cst.VERSION, cst.SPIDER)
check_data = dict(
    project=cst.PROJECT,
    _version=cst.VERSION,
    spider=cst.SPIDER,
    trigger='cron',  # if not request.form.get('trigger'): return
    hour='3'
)
run_data = {
    '1': 'on',
    '2': 'on',
    '3': 'on',  # It's out of range
    'checked_amount': '3',
    'filename': FILENAME
}
run_data_single_scrapyd = {
    '1': 'on',
    'checked_amount': '1',
    'filename': FILENAME
}


def test_check_with_task(app, client):
    req(app, client, view='schedule.check', kws=dict(node=NODE), data=check_data,
        jskws=dict(cmd="-d _version=%s" % cst.VERSION, filename=FILENAME))


def test_run_with_task(app, client):
    # ScrapydWeb_demo.egg: custom_settings = {}, also output specific settings & arguments in the log
    upload_file_deploy(app, client, filename='ScrapydWeb_demo_no_request.egg', project=cst.PROJECT,
                       redirect_project=cst.PROJECT)

    req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='enable'), ins='STATE_RUNNING', nos='STATE_PAUSED')

    with app.test_request_context():
        text, __ = req(app, client, view='schedule.run', kws=dict(node=NODE), data=run_data,
                       location=url_for('tasks', node=NODE))
    m = re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text))
    task_id = int(m.group(1))
    print("task_id: %s" % task_id)
    metadata['task_id'] = task_id

    __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert js['data']['selected_nodes'] == [1, 2]


def test_check_result(app, client):
    task_id = metadata['task_id']
    sleep(2)
    # The first execution has not finished yet: self.sleep_seconds_before_retry = 3
    req(app, client, view='tasks', kws=dict(node=NODE),
        ins=["id: %s," % task_id, "prev_run_result: 'FAIL 0, PASS 0',", "fail_times: 0,", "run_times: 1,"])
    text, __ = req(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
                   ins=["fail_count: 0,", "pass_count: 0,", ":total='1'"])
    with app.test_request_context():
        url_delete = url_for('tasks.xhr', node=NODE, action='delete', task_id=task_id)
    # in the task results page: url_action: '/1/tasks/xhr/delete/5/10/',
    task_result_id = int(re.search(r'%s(\d+)/' % url_delete, text).group(1))
    print("task_result_id: %s" % task_result_id)
    metadata['task_result_id'] = task_result_id
    with app.test_request_context():
        url_delete_task_result = url_for('tasks.xhr', node=NODE, action='delete',
                                         task_id=task_id, task_result_id=task_result_id)
    assert url_delete_task_result in text
    sleep(8)
    req(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id, task_result_id=task_result_id),
        ins=["node: 1,", "server: '%s'," % app.config['SCRAPYD_SERVERS'][0],
             "status_code: 200,", "status: 'ok',"])  # , ":total='1'"

    sleep(20)
    # The first execution has finished
    req(app, client, view='tasks', kws=dict(node=NODE),
        ins=["id: %s," % task_id, "prev_run_result: 'FAIL 1, PASS 1',", "fail_times: 1,", "run_times: 'FAIL 1 / 1',"])
    req(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
        ins=["fail_count: 1,", "pass_count: 1,", ":total='1'"])
    req(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id, task_result_id=task_result_id),
        ins=["node: 1,", "server: '%s'," % app.config['SCRAPYD_SERVERS'][0], "status_code: 200,", "status: 'ok',",
             "node: 2,", "server: '%s'," % app.config['SCRAPYD_SERVERS'][-1], "status_code: -1,", "status: 'error',",
             ":total='2'"])
    __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert '03:00:00' in js['data']['apscheduler_job']['next_run_time']


def test_edit_task(app, client):
    task_id = metadata['task_id']
    # http://127.0.0.1:5000/1/schedule/?task_id=1
    req(app, client, view='schedule', kws=dict(node=NODE, task_id=task_id),
        ins=["checked />[1] %s" % app.config['SCRAPYD_SERVERS'][0],
             "checked />[2] %s" % app.config['SCRAPYD_SERVERS'][-1]])

    check_data_ = dict(check_data)
    check_data_.update(task_id=task_id, hour='6')
    req(app, client, view='schedule.check', kws=dict(node=NODE), data=check_data_,
        jskws=dict(cmd="-d _version=%s" % cst.VERSION, filename=FILENAME))

    with app.test_request_context():
        metadata['location'] = url_for('tasks', node=NODE)
    text, __ = req(app, client, view='schedule.run', kws=dict(node=NODE), data=run_data_single_scrapyd,
                   location=metadata['location'])
    m = re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text))
    assert int(m.group(1)) == task_id

    __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert js['data']['selected_nodes'] == [1]

    sleep()
    req(app, client, view='tasks', kws=dict(node=NODE),
        ins=["fail_times: 1,", "run_times: 'FAIL 1 / 2',"])
    text, __ = req(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
                   ins=["fail_count: 0,", "fail_count: 1,", "pass_count: 1,", ":total='2'"])
    with app.test_request_context():
        url_delete = url_for('tasks.xhr', node=NODE, action='delete', task_id=task_id)
    # in the task results page: url_action: '/1/tasks/xhr/delete/5/10/',
    new_task_result_id = int(re.search(r'%s(\d+)/' % url_delete, text).group(1))
    print("new_task_result_id: %s" % new_task_result_id)
    req(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id, task_result_id=new_task_result_id),
        ins=["node: 1,", "server: '%s'," % app.config['SCRAPYD_SERVERS'][0],
             "status_code: 200,", "status: 'ok',", ":total='1'"])

    __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert '06:00:00' in js['data']['apscheduler_job']['next_run_time']

    req(app, client, view='schedule', kws=dict(node=NODE, task_id=task_id),
        ins="checked />[1] %s" % app.config['SCRAPYD_SERVERS'][0],
        nos="checked />[2] %s" % app.config['SCRAPYD_SERVERS'][-1])


# ['selected_nodes'] == [1] in test_edit_task() above
# switch between task_results.html and task_results_with_job.html
def test_switch_template(app, client):
    task_id = metadata['task_id']
    task_result_id = metadata['task_result_id']
    req(app, client, view='tasks.xhr',
        kws=dict(node=NODE, action='delete', task_id=task_id, task_result_id=task_result_id))
    req(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
        ins=["status_code: 200,", "status: 'ok',", ":total='1'"],
        nos=["status_code: -1,", "status: 'error',", 'label="Fail count"', 'label="Server"'])

    switch_scrapyd(app)

    req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='fire', task_id=task_id))
    sleep(2)
    req(app, client, view='tasks', kws=dict(node=NODE),
        ins=["id: %s," % task_id, "prev_run_result: 'FAIL 0, PASS 0',", "run_times: 2,"])
    req(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
        ins=['label="Fail count"', "pass_count: 0,", "fail_count: 0,", "pass_count: 1,", ":total='2'"],
        nos=['label="Server"', "status_code:", "status:"])

    sleep(28)
    req(app, client, view='tasks', kws=dict(node=NODE),
        ins=["id: %s," % task_id, "prev_run_result: 'FAIL 1, PASS 0',", "run_times: 'FAIL 1 / 2',"])
    req(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
        ins=["status_code: 200,", "status: 'ok',", "status_code: -1,", "status: 'error',", ":total='2'"],
        nos=['label="Fail count"', 'label="Server"'])

    req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='delete', task_id=task_id))


# 1: check and run
# 9: start_time
# 10: first execute
# 11: first check result
# 15: final execute
# 19: end_time
# 21: final check result
def test_task_start_execute_end(app, client):
    while True:
        now_datetime = datetime.now()
        if now_datetime.second % 10 != 1:
            sleep(1)
        else:
            break
    start_datetime = now_datetime + timedelta(seconds=8)
    first_execute_datetime = now_datetime + timedelta(seconds=9)
    second_execute_datetime = now_datetime + timedelta(seconds=14)
    end_datetime = now_datetime + timedelta(seconds=18)
    check_data_ = dict(check_data)
    check_data_.update(action='add', hour='*', minute='*', second='*/5',
                       start_date=start_datetime.strftime("%Y-%m-%d %H:%M:%S"),
                       end_date=end_datetime.strftime("%Y-%m-%d %H:%M:%S"))
    req(app, client, view='schedule.check', kws=dict(node=NODE), data=check_data_,
        jskws=dict(cmd="-d _version=%s" % cst.VERSION, filename=FILENAME))
    text, __ = req(app, client, view='schedule.run', kws=dict(node=NODE), data=run_data_single_scrapyd,
                   location=metadata['location'])
    m = re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text))
    task_id = int(m.group(1))
    print("task_id: %s" % task_id)
    with app.test_request_context():
        url_pause = url_for('tasks.xhr', node=NODE, action='pause', task_id=task_id)
        url_resume = url_for('tasks.xhr', node=NODE, action='resume', task_id=task_id)
        url_delete = url_for('tasks.xhr', node=NODE, action='delete', task_id=task_id)
        url_task_results = url_for('tasks', node=NODE, task_id=task_id)
    req(app, client, view='tasks', kws=dict(node=NODE),
        ins=[url_pause, url_task_results,
             "id: %s," % task_id, "prev_run_result: '%s'," % cst.NA, "run_times: 0,"],
        nos=[url_resume, url_delete])
    __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert first_execute_datetime.strftime("%Y-%m-%d %H:%M:%S") in js['data']['apscheduler_job']['next_run_time']

    sleep(10)
    # The first execution may or may not has finished
    req(app, client, view='tasks', kws=dict(node=NODE), ins=["id: %s," % task_id, "run_times: 1,"])
    req(app, client, view='tasks', kws=dict(node=NODE),
        ins=[url_pause, url_task_results, "id: %s," % task_id, "run_times: 1,"],
        nos=[url_resume, url_delete])
    req(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id), ins=":total='1'")
    __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert second_execute_datetime.strftime("%Y-%m-%d %H:%M:%S") in js['data']['apscheduler_job']['next_run_time']

    sleep(10)
    req(app, client, view='tasks', kws=dict(node=NODE), ins=["id: %s," % task_id, "run_times: 2,"])
    req(app, client, view='tasks', kws=dict(node=NODE),
        ins=[url_delete, url_task_results, "id: %s," % task_id, "next_run_time: '%s'," % cst.NA, "run_times: 2,"],
        nos=[url_pause, url_resume])
    req(app, client, view='tasks', kws=dict(node=NODE, task_id=task_id),
        ins=["status_code: 200,", "status: 'ok',", ":total='2'"],
        nos=["status_code: -1,", "status: 'error',"])
    __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert js['data']['apscheduler_job'] is None

    req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='delete', task_id=task_id))


# Visit Timer Tasks: remove_apscheduler_job_without_task()
# execute_task():   if not task: apscheduler_job.remove()
def test_auto_remove_apscheduler_job_if_task_not_exist(app, client):
    check_data_ = dict(check_data)
    check_data_.update(action='add')

    for kind in ['visit timer tasks', 'execute_task()']:
        req(app, client, view='schedule.check', kws=dict(node=NODE), data=dict(check_data_),
            jskws=dict(cmd="-d _version=%s" % cst.VERSION, filename=FILENAME))
        text, __ = req(app, client, view='schedule.run', kws=dict(node=NODE), data=run_data_single_scrapyd,
                       location=metadata['location'])
        m = re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text))
        task_id = int(m.group(1))
        print("task_id: %s" % task_id)

        __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
        assert '03:00:00' in js['data']['apscheduler_job']['next_run_time']

        req(app, client, view='tasks.xhr',
            kws=dict(node=NODE, action='delete', task_id=task_id, ignore_apscheduler_job='True'))

        __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id),
                     jskws=dict(message="apscheduler_job #{id} found. Task #{id} not found".format(id=task_id)))
        assert js['data']['apscheduler_job'] == task_id

        # apscheduler_job #1 removed since task #1 not exist
        if kind == 'execute_task()':
            req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='fire', task_id=task_id))
            sleep()
        else:
            req(app, client, view='tasks', kws=dict(node=NODE),
                ins="apscheduler_job #{id} removed since task #{id} not exist".format(id=task_id),
                nos="id: %s," % task_id)
        __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id),
                     jskws=dict(
                         status=cst.ERROR,
                         message="apscheduler_job #{id} not found. Task #{id} not found".format(id=task_id)))
        assert js['data'] is None


def test_execute_task_exception(app, client):
    check_data_ = dict(check_data)
    check_data_.update(action='add')

    req(app, client, view='schedule.check', kws=dict(node=NODE), data=check_data_,
        jskws=dict(cmd="-d _version=%s" % cst.VERSION, filename=FILENAME))

    with app.test_request_context():
        text, __ = req(app, client, view='schedule.run', kws=dict(node=NODE), data=run_data,
                       location=url_for('tasks', node=NODE))
    m = re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text))
    task_id = int(m.group(1))
    print("task_id: %s" % task_id)

    __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
    assert js['data']['selected_nodes'] == [1, 2]

    # req_single_scrapyd would set single_scrapyd=True
    req_single_scrapyd(app, client, view='tasks.xhr', kws=dict(node=1, action='fire', task_id=task_id))

    sleep()

    req(app, client, view='tasks', kws=dict(node=1),
        ins=["id: %s," % task_id, "prev_run_result: 'FAIL 1, PASS 1',", "fail_times: 1,", "run_times: 'FAIL 1 / 1',"])
    text, __ = req(app, client, view='tasks', kws=dict(node=1, task_id=task_id),
                   ins=["fail_count: 1,", "pass_count: 1,", ":total='1'"])
    with app.test_request_context():
        url_delete = url_for('tasks.xhr', node=1, action='delete', task_id=task_id)
    # in the task results page: url_action: '/1/tasks/xhr/delete/5/10/',
    task_result_id = int(re.search(r'%s(\d+)/' % url_delete, text).group(1))
    print("task_result_id: %s" % task_result_id)
    # In baseview.py: assert 0 < self.node <= self.SCRAPYD_SERVERS_AMOUNT
    # Note that AssertionError would be raise directly in test, whereas internal_server_error() would return 500.html
    # instead when the app is actually running, getting '500 error node index error: 2, which should be between 1 and 1'
    req(app, client, view='tasks', kws=dict(node=1, task_id=task_id, task_result_id=task_result_id),
        ins=["node: 1,", "server: '%s'," % app.config['SCRAPYD_SERVERS'][0], "status_code: 200,", "status: 'ok',",
             "node: 2,", "status_code: -1,", "status: 'exception',", "node index error", ":total='2'"])

    req(app, client, view='tasks.xhr', kws=dict(node=1, action='delete', task_id=task_id))


def test_delete_task_or_task_result_on_the_fly(app, client):
    for kind in ['delete_task', 'delete_task_result']:
        check_data_ = dict(check_data)

        req(app, client, view='schedule.check', kws=dict(node=NODE), data=check_data_,
            jskws=dict(cmd="-d _version=%s" % cst.VERSION, filename=FILENAME))

        with app.test_request_context():
            text, __ = req(app, client, view='schedule.run', kws=dict(node=NODE), data=run_data,
                           location=url_for('tasks', node=NODE))
        m = re.search(cst.TASK_NEXT_RUN_TIME_PATTERN, unquote_plus(text))
        task_id = int(m.group(1))
        print("task_id: %s" % task_id)

        __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='dump', task_id=task_id))
        assert js['data']['selected_nodes'] == [1, 2]

        sleep(2)
        # the first execution has not finished yet
        __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list', task_id=task_id))
        assert len(js['ids']) == 1
        task_result_id = js['ids'][0]
        __, js = req(app, client, view='tasks.xhr',
                     kws=dict(node=NODE, action='list', task_id=task_id, task_result_id=task_result_id))
        assert len(js['ids']) == 1

        if kind == 'delete_task':
            req(app, client, view='tasks.xhr',
                kws=dict(node=NODE, action='delete', task_id=task_id))
        else:
            req(app, client, view='tasks.xhr',
                kws=dict(node=NODE, action='delete', task_id=task_id, task_result_id=task_result_id))

        __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list'))
        if kind == 'delete_task':
            assert task_id not in js['ids']
        else:
            assert task_id in js['ids']

        __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list', task_id=task_id))
        assert len(js['ids']) == 0
        __, js = req(app, client, view='tasks.xhr',
                     kws=dict(node=NODE, action='list', task_id=task_id, task_result_id=task_result_id))
        assert len(js['ids']) == 0

        sleep(28)
        req(app, client, view='tasks.xhr',
            kws=dict(node=NODE, action='delete', task_id=task_id, task_result_id=task_result_id))
        __, js = req(app, client, view='tasks.xhr', kws=dict(node=NODE, action='list', task_id=task_id))
        assert len(js['ids']) == 0
        __, js = req(app, client, view='tasks.xhr',
                     kws=dict(node=NODE, action='list', task_id=task_id, task_result_id=task_result_id))
        assert len(js['ids']) == 0

        req(app, client, view='tasks.xhr', kws=dict(node=1, action='delete', task_id=task_id))



# coding: utf-8
import json

from tests.utils import cst, req, sleep


def test_send_text_page(app, client):
    req(app, client, view='sendtext', kws=dict(node=1), ins=['Send text via', 'import scrapy'])


def test_email_pass(app, client):
    if not (app.config.get('ENABLE_EMAIL_ALERT', True) and app.config.get('EMAIL_PASSWORD', '')):
        print("ENABLE_EMAIL_ALERT False or EMAIL_PASSWORD empty, skip test.")
        return

    def check_pass(recipients=None, subject='Email from #scrapydweb', text=None):
        assert js['status'] == cst.OK
        assert js['result']['reason'] == 'Sent'
        assert js['result']['sender'] == app.config['EMAIL_SENDER']
        if recipients is not None:
            assert js['result']['recipients'] == recipients
        if subject is not None:
            assert js['result']['subject'] == subject
        if text is not None:
            assert js['result']['text'] == text
        assert 'debug' not in js
        assert js['when']
        sleep(10)

    # 'email'
    nos = ['debug', 'email_password']
    __, js = req(app, client, view='sendtextapi', kws=dict(opt='email'), nos=nos)
    check_pass(text='test')

    # 'email/<text>'
    text = 'test email text'
    __, js = req(app, client, view='sendtextapi', kws=dict(opt='email', text=text))
    check_pass(text=text)

    # 'email/<channel_chatid_subject>/<text>'
    text = 'test email subject text'
    channel_chatid_subject = text + ' #scrapydweb'
    kws = dict(opt='email', channel_chatid_subject=channel_chatid_subject, text=text)
    __, js = req(app, client, view='sendtextapi', kws=kws)
    check_pass(text=text, subject=channel_chatid_subject)

    # ?recipients=&subject=
    recipients = app.config['EMAIL_RECIPIENTS'] + ['my8100@gmail.com']
    text = 'test email text recipients and subject in query strings'
    subject = text + ' #scrapydweb'
    kws = dict(opt='email', text=text, recipients=';'.join(recipients), subject=subject)
    __, js = req(app, client, view='sendtextapi', kws=kws)
    check_pass(recipients=recipients, subject=subject, text=text)

    # post form
    recipients = ['my8100@gmail.com'] + app.config['EMAIL_RECIPIENTS']
    value = 'test email post form #scrapydweb'
    subject = value + ' #scrapydweb'
    # Cannot directly pass in a list type to the data, otherwise, only the first recipient could be recognized.
    data = dict(recipients=','.join(recipients), subject=subject, a=1, arg=value, Chinese=u'中文')
    __, js = req(app, client, view='sendtextapi', kws=dict(opt='email'), data=data)
    check_pass(recipients=recipients, subject=subject, text=None)
    data['a'] = str(data['a'])
    assert json.loads(js['result']['text']) == data

    # post json
    # requests.post('http://httpbin.org/post', data=dict(a=1, b=2)).json()
    # {'data': '',
     # 'form': {'a': '1', 'b': '2'},
     # 'headers': {'Content-Type': 'application/x-www-form-urlencoded',},
     # 'json': None,}
    # requests.post('http://httpbin.org/post', json=dict(a=1, b=2)).json()
    # {'data': '{"a": 1, "b": 2}',
     # 'form': {},
     # 'headers': {'Content-Type': 'application/json',},
     # 'json': {'a': 1, 'b': 2},}
    # https://stackoverflow.com/questions/10434599/get-the-data-received-in-a-flask-request
    # request.headers:
    # Content-Type: application/x-www-form-urlencoded
    # Content-Type: application/json
    # https://stackoverflow.com/questions/28836893/how-to-send-requests-with-jsons-in-unit-tests
    recipients = app.config['EMAIL_RECIPIENTS'] + ['my8100@gmail.com']
    value = 'test email post json #scrapydweb'
    subject = value + ' #scrapydweb'
    data = dict(recipients=recipients, subject=subject, b=2, arg=value, Chinese=u'中文')
    __, js = req(app, client, view='sendtextapi', kws=dict(opt='email'),
                 data=json.dumps(data), content_type='application/json')
    check_pass(recipients=recipients, subject=subject, text=None)
    assert json.loads(js['result']['text']) == data


def test_email_fail(app, client):
    # EMAIL_RECIPIENTS invalid
    # "reason": "{'1@2': (501, b'Bad address syntax')}"
    try:
        req(app, client, view='sendtextapi', kws=dict(opt='email', recipients='1@invalid_recipients.com'),
            jskws=dict(status=cst.ERROR), jskeys=['status', 'result', 'debug', 'when'],
            ins='email_password', nos='Sent')
    except AssertionError:
        # Chances are that the email may be sent successfully.
        pass

    # EMAIL_PASSWORD unset
    app.config['EMAIL_PASSWORD'] = ''
    req(app, client, view='sendtextapi', kws=dict(opt='email'),
        jskws=dict(status=cst.ERROR, result="The EMAIL_PASSWORD option is unset"))

    # EMAIL_PASSWORD invalid
    # "reason": "Error: 请使用授权码登录。详情请看: http://service.mail.qq.com/cgi-bin/help?subtype=1&&id=28&&no=1001256"
    app.config['EMAIL_PASSWORD'] = 'fake_password'
    req(app, client, view='sendtextapi', kws=dict(opt='email'),
        jskws=dict(status=cst.ERROR), jskeys=['status', 'result', 'debug', 'when'],
        ins='email_password', nos='Sent')


def test_slack_pass(app, client):
    if not app.config['SLACK_TOKEN']:
        print("SLACK_TOKEN unset")
        return

    def check_pass(text=None):
        assert js['url'] == 'https://slack.com/api/chat.postMessage'
        assert js['status_code'] == 200
        assert js['status'] == cst.OK
        assert js['result']['ok'] is True
        if text is not None:
            assert js['result']['message']['text'] == text
        assert 'debug' not in js
        assert js['when']

    # 'slack'
    nos = ['debug', 'token', 'error']
    __, js = req(app, client, view='sendtextapi', kws=dict(opt='slack'), nos=nos)
    check_pass('test')
    channel_general = js['result']['channel']

    # 'slack/<text>'
    text = 'test slack text'
    __, js = req(app, client, view='sendtextapi', kws=dict(opt='slack', text=text))
    check_pass(text)

    # 'slack/<channel_chatid_subject>/<text>'
    channel = 'random'  # general -> "channel": "CLX124N0Z", random -> "channel": "CLRUUNCCT"
    text = 'test slack channel %s and text' % channel
    kws = dict(opt='slack', channel_chatid_subject=channel, text=text)
    __, js = req(app, client, view='sendtextapi', kws=kws)
    check_pass(text)
    channel_random = js['result']['channel']
    assert channel_random != channel_general

    # post form and channel in query string
    channel = 'random'
    value = "slack post form and channel %s in query string" % channel
    data = dict(a=1, arg=value, Chinese=u'中文')
    __, js = req(app, client, view='sendtextapi', kws=dict(opt='slack', channel=channel), data=data)
    check_pass()
    data['a'] = str(data['a'])
    assert json.loads(js['result']['message']['text']) == data
    assert js['result']['channel'] == channel_random

    # post json
    channel = 'random'
    value = "slack post json and channel %s in data" % channel
    data = dict(channel=channel, b=2, arg=value, Chinese=u'中文')
    __, js = req(app, client, view='sendtextapi', kws=dict(opt='slack'),
                 data=json.dumps(data), content_type='application/json')
    check_pass()
    assert json.loads(js['result']['message']['text']) == data
    assert js['result']['channel'] == channel_random


def test_slack_fail(app, client):
    kws = dict(opt='slack')
    jskeys = ['url', 'status_code', 'status', 'result', 'debug', 'when']

    def check_fail(status_code=None, result=None):
        assert js['status'] == 'error'
        assert js['url'] == 'https://slack.com/api/chat.postMessage'
        if status_code is not None:
            assert js['status_code'] == status_code
        if result is not None:
            assert js['result'] == result
        assert js['when']

    # SLACK_TOKEN unset
    token = app.config['SLACK_TOKEN']
    app.config['SLACK_TOKEN'] = ''
    req(app, client, view='sendtextapi', kws=kws,
        jskws=dict(status=cst.ERROR, result="The SLACK_TOKEN option is unset"))

    if not token:
        return
    else:
        app.config['SLACK_TOKEN'] = token

    # SLACK_TOKEN invalid
    fake_token = 'fake_token'
    app.config['SLACK_TOKEN'] = fake_token
    __, js = req(app, client, view='sendtextapi', kws=kws, jskeys=jskeys)
    check_fail(status_code=200, result={'error': 'invalid_auth', 'ok': False})
    assert js['debug'] == {'channel': app.config['SLACK_CHANNEL'], 'text': 'test', 'token': fake_token}

    app.config['SLACK_TOKEN'] = token

    # SLACK_CHANNEL invalid
    fake_channel = 'fake_channel'
    app.config['SLACK_CHANNEL'] = fake_channel
    __, js = req(app, client, view='sendtextapi', kws=kws, jskeys=jskeys)
    check_fail(status_code=200, result={'error': 'channel_not_found', 'ok': False})
    assert js['debug'] == {'channel': fake_channel, 'text': 'test', 'token': app.config['SLACK_TOKEN']}


# On win7, set ENABLE_TELEGRAM_ALERT=True to test Telegram seperately
def test_telegram_pass(app, client):
    if not (app.config['TELEGRAM_TOKEN'] and app.config['ENABLE_TELEGRAM_ALERT']):
        print("ENABLE_TELEGRAM_ALERT False or TELEGRAM_TOKEN unset")
        return

    def check_pass(text=None):
        assert js['url'] == 'https://api.telegram.org/bot%s/sendMessage' % app.config['TELEGRAM_TOKEN']
        assert js['status_code'] == 200
        assert js['status'] == cst.OK
        assert js['result']['ok'] is True
        if text is not None:
            assert js['result']['result']['text'] == text
        assert 'debug' not in js
        assert js['when']

    # 'telegram' | 'tg'
    nos = ['debug', 'token', 'error']
    __, js = req(app, client, view='sendtextapi', kws=dict(opt='telegram'), nos=nos)
    check_pass('test')
    __, js = req(app, client, view='sendtextapi', kws=dict(opt='tg'), nos=nos)
    check_pass('test')

    # 'tg/<text>'
    text = 'test tg text'
    __, js = req(app, client, view='sendtextapi', kws=dict(opt='tg', text=text))
    check_pass(text)

    # 'tg/<channel_chatid_subject>/<text>'
    chat_id = app.config['TELEGRAM_CHAT_ID']
    app.config['TELEGRAM_CHAT_ID'] = 0
    text = 'test tg chat_id text'
    kws = dict(opt='tg', channel_chatid_subject=chat_id, text=text)
    __, js = req(app, client, view='sendtextapi', kws=kws)
    check_pass(text)

    # post json
    value = "tg post json"
    data = dict(chat_id=chat_id, b=2, arg=value, Chinese=u'中文')
    __, js = req(app, client, view='sendtextapi', kws=dict(opt='tg'),
                 data=json.dumps(data), content_type='application/json')
    check_pass()
    assert json.loads(js['result']['result']['text']) == data

    app.config['TELEGRAM_CHAT_ID'] = chat_id

    # post form
    value = "tg post form"
    data = dict(a=1, arg=value, Chinese=u'中文')
    __, js = req(app, client, view='sendtextapi', kws=dict(opt='tg'), data=data)
    check_pass()
    data['a'] = str(data['a'])
    assert json.loads(js['result']['result']['text']) == data


# On win7, set ENABLE_TELEGRAM_ALERT=True to test Telegram seperately
def test_telegram_fail(app, client):
    kws = dict(opt='telegram')

    def check_fail(status_code=None, result=None, token=app.config['TELEGRAM_TOKEN']):
        assert js['status'] == 'error'
        if status_code is not None:
            assert js['status_code'] == status_code
            assert js['url'] == 'https://api.telegram.org/bot%s/sendMessage' % token
        if result is not None:
            assert js['result'] == result
        assert js['when']

    # TELEGRAM_TOKEN unset
    token = app.config['TELEGRAM_TOKEN']
    app.config['TELEGRAM_TOKEN'] = ''
    req(app, client, view='sendtextapi', kws=kws,
        jskws=dict(status=cst.ERROR, result="The TELEGRAM_TOKEN option is unset"))

    if not (token and app.config['ENABLE_TELEGRAM_ALERT']):
        return
    else:
        app.config['TELEGRAM_TOKEN'] = token

    # TELEGRAM_TOKEN invalid
    fake_token = 'fake_token'
    app.config['TELEGRAM_TOKEN'] = fake_token
    __, js = req(app, client, view='sendtextapi', kws=kws)
    result = dict(description="Not Found", error_code=404, ok=False)
    check_fail(status_code=404, result=result, token=fake_token)
    assert js['debug'] == dict(token=fake_token, chat_id=app.config['TELEGRAM_CHAT_ID'], text='test')

    fake_token_ = token + '_'
    app.config['TELEGRAM_TOKEN'] = fake_token_
    __, js = req(app, client, view='sendtextapi', kws=kws)
    result = dict(description="Unauthorized", error_code=401, ok=False)
    check_fail(status_code=401, result=result, token=fake_token_)
    assert js['debug'] == dict(token=fake_token_, chat_id=app.config['TELEGRAM_CHAT_ID'], text='test')

    app.config['TELEGRAM_TOKEN'] = token

    # TELEGRAM_CHAT_ID invalid
    chat_id = app.config['TELEGRAM_CHAT_ID']
    fake_chat_id = 0
    app.config['TELEGRAM_CHAT_ID'] = fake_chat_id
    __, js = req(app, client, view='sendtextapi', kws=kws)
    check_fail(status_code=400, result=dict(description="Bad Request: chat not found", error_code=400, ok=False))
    assert js['debug'] == dict(token=token, chat_id=fake_chat_id, text='test')

    fake_chat_id_ = chat_id - 1
    app.config['TELEGRAM_CHAT_ID'] = fake_chat_id_
    __, js = req(app, client, view='sendtextapi', kws=kws)
    check_fail(status_code=400, result=dict(description="Bad Request: chat not found", error_code=400, ok=False))
    assert js['debug'] == dict(token=token, chat_id=fake_chat_id_, text='test')
